#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
from __future__ import division
from builtins import range
from past.utils import old_div
import copy
import numpy as np
import scipy.stats
import matplotlib as mpl
import matplotlib.pyplot as plt
try:
	from pandas.plotting import lag_plot # py3.6
except ImportError:
	from pandas.tools.plotting import lag_plot # py3.5
import seaborn as sns
sns.set_style("ticks")

from matplotlib.patches import RegularPolygon
from string import ascii_uppercase


def plot_wells(df, x='Time', y='OD', plot_func=plt.plot, output_filename=None):
	"""Plot a grid of plots, one for each well in the plate.

	The facetting is done by the ``Row`` and ``Col`` columns of `df`.
	The colors are given by the ``Color`` column, 
	the labels of the colors are given by the ``Strain`` column.
	If ``Strain`` is missing then the coloring is done by the ``Well`` column.

	Parameters
	----------
	df : pandas.DataFrame
		growth curve data, see :py:mod:`curveball.ioutils` for a detailed definition.
	x : str, optional
		name of column for x-axis, defaults to ``Time``.
	y : str, optional
		name of column for y-axis, defaults to ``OD``.
	plot_func : func, optional
		function to use for plotting, defaults to :py:func:`matplotlib.pyplot.plot`
	output_filename : str, optional 
		filename to save the resulting figure; if not given, figure is not saved.

	Returns
	-------
	seaborn.FacetGrid
		figure object.
	"""
	if 'Strain' in df:
		hue = 'Strain'
		palette = df.Color.unique() if 'Color' in df else sns.color_palette()
		hue_order = df.Strain.unique()
		palette[palette == '#ffffff'] = '#000000'
	else:
		hue = 'Well'
		palette = sns.color_palette()
		hue_order = df.Well
	height = len(df.Row.unique())
	width = len(df.Col.unique())
	g = sns.FacetGrid(df, hue=hue, col='Col', row='Row',
                      palette=palette, hue_order=hue_order,
                      sharex=True, sharey=True, height=1,
                      aspect=old_div(width,float(height)), despine=True,margin_titles=True)
	g.map(plot_func, x, y)
	g.fig.set_figwidth(width)
	g.fig.set_figheight(height)
	plt.locator_params(nbins=4) # 4 ticks is enough
	g.set_axis_labels('','') 	# remove facets axis labels
	g.fig.text(0.5, 0, x, size='x-large') # xlabel
	g.fig.text(-0.01, 0.5, y, size='x-large', rotation='vertical') # ylabel
	if output_filename:
		g.savefig(output_filename, bbox_inches='tight', pad_inches=1)
	return g


def plot_strains(data, x='Time', y='OD', plot_func=plt.plot, by=None, agg_func=np.mean, hue='Strain', color=None, output_filename=None, **kwargs):
	"""Aggregate by strain and plot the results on one figure with different color for each strain.

	The grouping of the data is done by the ``Strain`` and either ``Cycle Nr.`` or ``Time`` columns of `data`;
	the aggregation is done by the `agg_func`, which defaults to :py:func:`numpy.mean`.
	The colors are given by the ``Color`` column, the labels of the colors are given by the ``Strain`` column of `data`.

	Parameters
	----------
	data : pandas.DataFrame
		growth curve data, see :py:mod:`curveball.ioutils` for a detailed definition.
	x : str, optional
		name of column for x-axis, defaults to ``Time``.
	y : str, optional
		name of column for y-axis, defaults to ``OD``.
	plot_func : func, optional
		function to use for plotting, defaults to :py:func:`matplotlib.pyplot.plot`
	by : tuple of str, optional
		used for grouping the data, defaults to ``('Strain', 'Cycle Nr.')`` or ``('Strain', 'Time')``, whichever is available.
	plot_func : func, optional
		function to use for aggregating the data, defaults to :py:func:`numpy.mean`.
	color : seaborn color palette
		a seaborn color palette to use if there is no ``Color`` column; if not given, using the default palette.
	output_filename : str, optional 
		filename to save the resulting figure; if not given, figure is not saved.

	Returns
	-------
	seaborn.FacetGrid
		figure object.

	Raises
	------
	ValueError
		raised if `by` isn't set and `data` doesn't contain ``Strain`` and either ``Time`` or ``Cycle Nr.``.
	"""
	if 'Color' in data:
		palette = data.Color.unique()
		palette[palette == '#ffffff'] = '#000000'
	else:
		palette = color or sns.color_palette()
	if by is None:
		if 'Cycle Nr.' in data and 'Strain' in data:
			by = ['Strain', 'Cycle Nr.']
		elif 'Time' in data and 'Strain' in data:
			by = ['Strain', 'Time']
	else:
		raise ValueError("If by is not set then data must have column Strain and either Time or Cycle Nr.")
	grp = data.groupby(by=by)
	numeric_cols = [col for col in data.select_dtypes(include=[np.number]).columns if col not in by]
	agg_func_name = 'mean' if agg_func is np.mean else agg_func
	agg = grp[numeric_cols].aggregate(agg_func_name).reset_index()
	g = sns.FacetGrid(agg, hue=hue, height=5, aspect=1.5, palette=palette, hue_order=data[hue].unique())
	g.map(plot_func, x, y);
	g.add_legend()
	if output_filename:
		g.savefig(output_filename, bbox_inches='tight', pad_inches=1)
	return g


def tsplot(data, x='Time', y='OD', ci_level=95, ax=None, color=None, output_filename=None, **kwargs):
	"""Time series plot of the data by strain (if applicable) or well.

	The grouping of the data is done by the value of `x` and ``Strain``, if such a column exists in `data`; 
	otherwise it is done by `x` and ``Well``.
	The aggregation is done by :py:func:`seaborn.lineplot` which calculates the mean with a confidence interval.
	The colors are given by the ``Color`` column, the labels of the colors are given by the ``Strain`` column; 
	if ``Strain`` and ``Color`` don't exist in `data` then
	the function will use a default palette and color the lines by well.

	Parameters
	----------
	data : pandas.DataFrame
		growth curve data, see :py:mod:`curveball.ioutils` for a detailed definition.
	x : str, optional
		name of column for x-axis, defaults to ``Time``.
	y : str, optional
		name of column for y-axis, defaults to ``OD``.
	ci_level : int, optional
		confidence interval width in precent (0-100), defaults to 95.
	ax : matplotlib.axes.Axes, optional
		plot into this axes, if not given create a new figure.
	color : seaborn color palette
		a seaborn color palette to use if there is no ``Color`` column; if not given, using the default palette.
	output_filename : str, optional 
		filename to save the resulting figure; if not given, figure is not saved.

	Returns
	-------
	matplotlib.axes.Axes
		axes object
	"""
	if 'Strain' in data:
		condition = 'Strain'		
	else:
		condition = 'Well'
	if 'Color' in data:
		palette = data['Color'].unique()
		palette[palette == '#ffffff'] = '#000000'
	else: 
		palette = color or sns.color_palette()

	try:
		g = sns.lineplot(
			data=data,
			x=x,
			hue=condition,
			y=y,
			errorbar=("ci", ci_level),
			palette=list(palette),
			ax=ax
		)
	except TypeError:
		g = sns.lineplot(
			data=data,
			x=x,
			hue=condition,
			y=y,
			err_style='band',
			ci=ci_level,
			palette=list(palette),
			ax=ax
		)
	sns.despine()
	if output_filename:
		g.figure.savefig(output_filename, bbox_inches='tight', pad_inches=1)
	return g


def plot_plate(data, edge_color='#888888', output_filename=None):
	"""Plot of the plate color mapping.

	The function will plot the color mapping in `data`:
	a grid with enough columns and rows for the ``Col`` and ``Row`` columns in `data`,
	where the color of each grid cell given by the ``Color`` column.

	Parameters
	----------
	data : pandas.DataFrame
		growth curve data, see :py:mod:`curveball.ioutils` for a detailed definition.
	edge_color : str
		color hex string for the grid edges.

	Returns
	-------
	fig : matplotlib.figure.Figure
		figure object
	ax : numpy.ndarray
		array of axis objects.
	"""
	plate = data.pivot(index='Row', columns='Col', values='Color').to_numpy()
	height, width = plate.shape
	fig = plt.figure(figsize=((width + 2.0) / 3.0, (height + 2.0) / 3.0))
	ax = fig.add_axes((0.05, 0.05, 0.9, 0.9),
	                            aspect='equal', frameon=False,
	                            xlim=(-0.05, width + 0.05),
	                            ylim=(-0.05, height + 0.05))
	for axis in (ax.xaxis, ax.yaxis):
	    axis.set_major_formatter(plt.NullFormatter())
	    axis.set_major_locator(plt.NullLocator())

	# Create the grid of squares
	squares = np.array([[RegularPolygon((i + 0.5, j + 0.5),
	                                         numVertices=4,
	                                         radius=0.5 * np.sqrt(2),
	                                         orientation=old_div(np.pi, 4),
	                                         ec=edge_color,
	                                         fc=plate[height-1-j,i])
	                          for j in range(height)]
	                         for i in range(width)])
	[ax.add_patch(sq) for sq in squares.flat]
	ax.set_xticks(np.arange(width) + 0.5)
	ax.set_xticklabels(np.arange(1, 1 + width))
	ax.set_yticks(np.arange(height) + 0.5)
	ax.set_yticklabels(ascii_uppercase[height-1::-1])
	ax.xaxis.tick_top()
	ax.yaxis.tick_left()
	ax.tick_params(length=0, width=0)
	if output_filename:
		fig.savefig(output_filename, bbox_inches='tight', pad_inches=1)
	return fig, ax


def plot_params_distribution(param_samples, color='k', cmap="viridis", alpha=None):
	"""Plots a distribution of model parameter samples generated with :py:func:`curveball.models.sample_params`.

	Parameters
	----------
	param_samples : pandas.DataFrame
		data frame of samples; each row is one sample, each column is one parameter.
	alpha : float
		transparency of plot markers, defaults to :math:`1/n^{1/4}` where *n* is number of rows in `param_samples`.

	Returns
	-------
	seaborn.Grid
		figure object
	"""
	nsamples = param_samples.shape[0]
	g = sns.PairGrid(param_samples)
	if alpha is None:
		alpha = 1.0 / np.power(nsamples, 1.0 / 4.0)
	g.map_upper(plt.scatter, alpha=alpha, color=color)
	g.map_lower(sns.kdeplot, cmap=cmap, legend=False, fill=True, thresh=0.05)
	g.map_diag(plt.hist, facecolor=color) # https://github.com/mwaskom/seaborn/pull/788
	return g


def _plot_fitted_histogram(data, rv=scipy.stats.norm, color='k', label=None, alpha=0.5, ax=None):
	"""This is basically `sns.distplot(fit=rv)`.
	TODO: `low,high = np.percentile(x, 2.5), np.percentile(x, 97.5)`
	"""
	if ax is None:
		fig, ax = plt.subplots(1, 1)
	else:
		fig = ax.figure
	rv_params = rv.fit(data)
	rv_inst = rv(*rv_params)	
	nbins = min(100, len(data))
	n, bins, patches = ax.hist(data, bins=nbins, color=color, alpha=alpha, density=True)
	ax.plot(bins, rv_inst.pdf(bins), color='k', lw=2)
	ax.annotate(
		r'$\mu={:.2g}, \sigma={:.2g}$'.format(rv_inst.mean(), 
		rv_inst.std()), 
		xy=(bins[len(bins)//2], np.max(n)), 
		xycoords="data", 
		horizontalalignment='center', 
		fontsize=plt.rcParams['axes.labelsize']
	)
	return fig, ax


def plot_model_residuals(model_fit, rv=scipy.stats.norm, color='k'):
	"""Plot of the residuals of a model fit.

	The function will plot the residuals - the difference between data and model - for a given model fit.
	The left panel shows the residuals over time; the right panel shows the histogram of the residuals with a fitted distribution curve.

	Parameters
	----------
	model_fit : lmfit.ModelResult
		the result of a model fitting procedure.
	rv : scipy.stats.rv_continuous, optional
		:py:class:`scipy.stats.rv_continuous` random variable whose probability density function (pdf)
		will be fitted to the histogram. Defaults to the normal distribution (`scipy.stats.norm`).
	color : str, optional
		color string for the plot, defaults to `k` for black.

	Returns
	-------
	fig : matplotlib.figure.Figure
		figure object
	ax : numpy.ndarray
		array of axis objects.
	"""
	w, h= plt.rcParams['figure.figsize']
	fig,ax = plt.subplots(1, 2, figsize=(w * 2, h))

	model_fit.plot_residuals(ax=ax[0], data_kws={'color': color}) # removed, causes bug in lmfit: fit_kws={'color': color})
	ax[0].set_xlabel('Time (hr)')
	ax[0].set_ylabel('Residuals')
	ax[0].legend().set_visible(False)
	ax[0].set_title('')

	_plot_fitted_histogram(model_fit.residual, rv=rv, color=color, ax=ax[1])
	ax[1].set(xlabel='Residuals', ylabel='Frequency')

	fig.tight_layout()
	sns.despine()	
	return fig, ax


def plot_residuals(df, time='Time', value='OD', resid_func=lambda x: x - x.mean(), rv=scipy.stats.norm, 
	color='k', ax=None):
	"""Plot of the residuals of in the data.

	The function will plot the residuals - the difference between data and average at each time point.
	The left panel shows the residuals over time.
	The middle panel shows the histogram of the residuals with a fitted distribution (defaults to Gaussian).
	The right panel shows the regression between the standard deviation at time `t+1` and `t` to identify autocorrelation.


	Parameters
	----------
	df : pandas.DataFrame
		a data frame with columns ``Time`` and ``OD``.
	time : str, optional
		name of column over which to group and plot the residuals. Defaults to ``Time``.
	value : str, optional
		name of column in `df` of the value on which to compute the residuals. Defaults to ``OD``.
	resid_func : function, optional
		function to calculate residuals. Defaults to ``x - x.mean()``.
	rv : scipy.stats.rv_continuous, optional
		:py:class:`scipy.stats.rv_continuous` random variable whose probability density function (pdf)
		will be fitted to the histogram. Defaults the normal distribution (:py:class:`scipy.stats.norm`).
	color : str, optional
		color string for the plot, defaults to `k` for black.

	Returns
	-------
	fig : matplotlib.figure.Figure
		figure object
	ax : numpy.ndarray
		array of axis objects.
	"""
	w, h= plt.rcParams['figure.figsize']
	fig,ax = plt.subplots(1, 3, figsize=(w * 3, h))

	residuals = df.groupby(time)[value].transform(resid_func).to_numpy()

	ax[0].plot(df[time], residuals, ls='', marker='o', color=color)
	ax[0].set(xlabel=time, ylabel='Residuals')	
	
	_plot_fitted_histogram(residuals, rv=rv, color=color, ax=ax[1])
	ax[1].set(xlabel='Residuals', ylabel='Frequency')
	
	sigmas = df.groupby(time)[value].std()	
	linreg = scipy.stats.linregress(sigmas.to_numpy()[:-1], sigmas.to_numpy()[1:])
	eq = r'$\sigma_{{t+1}} = {:.2g} + {:.2g} \sigma_{{t}}$'.format(linreg.intercept, linreg.slope)
	sigma_range = np.linspace(sigmas.min(), sigmas.max())
	ax[2].plot(sigma_range, sigma_range, color='k', ls='--', label=r'$\sigma_{t+1}=\sigma_{t}$')
	ax[2].plot(sigma_range, linreg.intercept + linreg.slope * sigma_range, color=color, label=eq)
	lag_plot(sigmas, c='k', ax=ax[2])
	ax[2].set(xlabel=r'$\sigma_{t}$', ylabel=r'$\sigma_{t+1}$')
	ax[2].legend(loc='upper left')

	fig.tight_layout()
	sns.despine()	
	return fig, ax


def plot_sample_fit(model_fit, param_samples, fit_kws=None, data_kws=None, sample_kws=None):
	"""Plot of sampled curve fits.

	The function will plot the main model fit and the sampled curve fits based on a table of sample parameters.


	Parameters
	----------
	model_fit : lmfit.ModelResult
		the result of a model fitting procedure.
	param_samples : pandas.DataFrame
		data frame of samples; each row is one sample, each column is one parameter.
	fit_kws, data_kws, sample_kws : dict
		dictionaries of plot directives for the fit, data, and sampled fit curves.

	Returns
	-------
	fig : matplotlib.figure.Figure
		figure object
	ax : numpy.ndarray
		array of axis objects.
	"""
	t = np.linspace(0, model_fit.userkws['t'].max())
	def f(params):
		return model_fit.model.eval(t=t, params=params)

	nsamples = param_samples.shape[0]
	_fit_kws = dict(linewidth=5)
	if fit_kws: _fit_kws.update(fit_kws)
	_data_kws = dict(marker='.')
	if data_kws: _data_kws.update(data_kws)
	_sample_kws = dict(linestyle='--', color='gray', alpha=1/np.sqrt(nsamples))
	if sample_kws: _sample_kws.update(sample_kws)

	ax = model_fit.plot_fit(init_kws={'ls': ''}, fit_kws=_fit_kws, data_kws=_data_kws)
	for i in range(nsamples):
		sample = param_samples.iloc[i, :]
		params = model_fit.params.copy()
		for k, v in params.items():
			if v.vary:
				params[k].set(value=sample[k])
		plt.plot(t, f(params), **_sample_kws)
	ax.legend().set_visible(False)
	ax.set(ylabel='OD', xlabel='Time', title='')

	sns.despine()
	return ax.figure, ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
from __future__ import print_function
from __future__ import division
import sys
import inspect
from warnings import warn
import numpy as np
from scipy.optimize import minimize
def derivative(func, x0, dx=1.0, n=1, order=3):
	"""Finite-difference derivative fallback for SciPy>=1.10 compatibility."""
	if n == 1:
		return (func(x0 + dx) - func(x0 - dx)) / (2.0 * dx)
	if n == 2:
		return (func(x0 + dx) - 2.0 * func(x0) + func(x0 - dx)) / (dx ** 2)
	return derivative(lambda t: derivative(func, t, dx, n - 1, order), x0, dx, 1, order)
import matplotlib.pyplot as plt
import pandas as pd
import lmfit
import sympy
import curveball
from curveball.utils import smooth


MIN_VALUES = {
	# based on nothing
	'y0': 1e-4,
	'K': 1e-4,
	'r': 1e-4,	
	'nu': 1e-1,
	'q0': 1e-4,
	'v': 1e-4
}

MAX_VALUES = {
	# based on min doubling time of ~5 minutes
	# r = 60 * log(2) / x where x is doubling time in minutes
	'r': 8,
	# based on nothing
	'nu': 10,
}


def lag(model_result=None, q0=None, v=None):
	r"""Calculate lag duration from model Parameters.

	Lag duration :math:`\lambda` is calculated as [Baranyi1994]_:

	.. math::

		\lambda = \frac{1}{v} \log{\Big( 1 + \frac{1}{q_0} \Big)}

	This definition follows from the Baranyi-Roberts model definition of :math:`A(t)`:

	.. math::

		A(t) - t = \frac{1}{v} \log{\Big(\frac{e^{-vt} + q_0}{1 + q_0} \Big)} \to_{t \to \infty} \lambda

	References
	----------
	.. [Baranyi1994] Baranyi, J., Roberts, T. A., 1994. `A dynamic approach to predicting bacterial growth in food <www.ncbi.nlm.nih.gov/pubmed/7873331>`_. Int. J. Food Microbiol.

	"""
	if model_result is not None:
		q0 = model_result.best_values.get('q0', np.inf)
		v = model_result.best_values.get('v', model_result.best_values['r'])
	elif q0 is None or v is None:
		raise ValueError("Either model_result or q0 and v should be given")
	if np.isinf(q0) or np.isinf(v):
		return 0.0
	return np.log(1.0 + 1.0 / q0) / v


def baranyi_roberts_function(t, y0, K, r, nu, q0, v):
	r"""The Baranyi-Roberts growth model is an extension of the logistic and Richards model that adds a lag phase [Baranyi1994]_.

	.. math::

		\frac{dy}{dt} = r \alpha(t) y \Big( 1 - \Big(\frac{y}{K}\Big)^{\nu} \Big) \Rightarrow

		y(t) = \frac{K}{\Big[1 - \Big(1 - \Big(\frac{K}{y_0}\Big)^{\nu}\Big) e^{-r \nu A(t)}\Big]^{1/\nu}}

		A(t) = \int_0^t{\alpha(s)ds} = \int_0^t{\frac{q_0}{q_0 + e^{-v s}} ds} = t + \frac{1}{v} \log{\Big( \frac{e^{-v t} + q0}{1 + q0} \Big)}


	- :math:`y_0`: initial population size
	- r: initial per capita growth rate
	- K: maximum population size
	- :math:`\nu`: curvature of the logsitic term
	- :math:`q_0`: initial physiological state of the population
	- v: rate of physiological state adjustment

	Parameters
	----------
	t : numpy.ndarray
		array of floats for time, usually in hours (:math:`t>0`)
	y0 : float
		initial population size (:math:`y_0>0`)
	r : float
		initial per capita growth rate
	K : float
		maximum population size (:math:`K>0`)
	nu : float
		curvature of the logsitic term (:math:`\nu>0`)
	q0 : float
		initial physiological state of the population (:math:`0<q_0<1`)
	v : float
		rate of physiological state adjustment (:math:`v>0`)

	Returns
	-------
	numpy.ndarray
		population size per time point in `t`.

	References
	----------
	.. [Baranyi1994] Baranyi, J., Roberts, T. A., 1994. `A dynamic approach to predicting bacterial growth in food <www.ncbi.nlm.nih.gov/pubmed/7873331>`_. Int. J. Food Microbiol.
	"""
	if np.isposinf(v):
		v = r
	if np.isposinf(q0):
		At = t
	else:
		At = t + (1.0 / v) * np.log((np.exp(-v * t) + q0) / (1.0 + q0))
	if np.isposinf(K):
		return y0 * np.exp(r * nu * At)
	else:
		return K / ((1 - (1 - (K / y0)**nu) * np.exp(-r * nu * At))**(1.0/nu))


def guess_nu(t, N, K=None, PLOT=False, PRINT=False):
	r"""Guesses the value of :math:`\nu` from the shape of the growth curve.

	Following [Richards1959]_:

	.. math::

		N_{max} = K (1 + \nu)^{-\frac{1}{\nu}}


	- :math:`N_{max}`: population size when the population growth rate (:math:`\frac{dN}{dt}`) is maximum
	- r: initial per capita growth rate 
	- K: maximum population size
	- :math:`\nu`: curvature of the logsitic term

	Parameters
	----------
	t : numpy.ndarray
		time
	N : numpy:ndarray
		`N[i]` is the population size at time `t[i]`
	K : float, optional
		a guess of `K`, the maximum population size. If not given, it is guessed.
	PLOT : bool, optional
		if :const:`True`, the function will plot the calculations.
	PRINT : bool, optional
		if :const:`True`, the function will print intermediate results of the calculations.

	Returns
	-------
	x : float
		the guess of :math:`\nu`.
	fig : matplotlib.figure.Figure
		if the argument `PLOT` was :const:`True`, the generated figure.
	ax : matplotlib.axes.Axes
		if the argument `PLOT` was :const:`True`, the generated axis.

	References
	----------
	.. [Richards1959] Richards, F. J. 1959. `A Flexible Growth Function for Empirical Use <http://dx.doi.org/10.1093/jxb/10.2.290>`_. Journal of Experimental Botany
	"""
	t, N = np.array(t), np.array(N)
	if K is None:
		K = N.max()
	# smooth and calculate derivative
	N_smooth = smooth(t, N) 	
	# calculate derivative	
	dNdt = derivative(N_smooth, t)
	assert len(t) > 0
	assert len(t) == len(dNdt)
	# only use the second part of the curve starting from N=K/e (inflexion point for nu=0)
	idx = N >= K * np.exp(-1)
	if idx.sum() < 10:
		warn('Less than 10 data points above K/e!')	
	t, N, dNdt = t[idx], N[idx], dNdt[idx]
	# find N at inflexion point
	i = dNdt.argmax()
	Ninf = float(np.asarray(N_smooth(t[i])).squeeze())
	# find nu that gives Ninf at inflexion point
	def target(nu):
		return np.abs(K * (1 + nu)**(-1.0 / nu)  - Ninf)
	opt_res = minimize(target, x0=1, bounds=[(0, None)])
	x = float(opt_res.x.item())
	y = target(x)
	y1 = target(1.0)
	
	if PRINT:
		print("f(1)=%.4f, f(%.4f)=%.4f" % (y1, x, y))
	if not opt_res.success and not np.allclose(y, 0):
		warn("Minimization warning in %s: %s\nGuessed nu=%.4f with f(nu)=%.4f" % (sys._getframe().f_code.co_name, opt_res.message, x, y))
	if y1 < y:
		warn("f(1)=%.4f < f(%.4f)=%.4f, Setting nu=1" % (y1, x, y))
		x = 1.0
	if PLOT:
		fs = plt.rcParams['figure.figsize']
		fig, ax = plt.subplots(1, 3, figsize=(fs[0] * 3, fs[1]))
		ax0, ax1, ax2 = ax
		
		N_t = N_smooth(t)
		ax0.plot(t, N_t, '-k')
		ax0.axvline(t[i], color='k', ls='--')
		ax0.axhline(N_t[i], color='k', ls='--')
		ax0.set_xlabel('Time')
		ax0.set_ylabel('N')
		
		ax1.plot(t, dNdt, '-k')
		ax1.axvline(t[i], color='k', ls='--')
		ax1.axhline(dNdt[i], color='k', ls='--')
		ax1.set_xlabel('Time')
		ax1.set_ylabel('dN/dt')
		dN = (dNdt.max() - dNdt.min()) / 10
		ax1.set_ylim(dNdt.min() - dN, dNdt.max() + dN)
		
		ax2.plot(np.logspace(-3,3), target(np.logspace(-3, 3)), 'k-')
		ax2.axvline(x, color='k', ls='--')
		ax2.axhline(y, color='k', ls='--')
		ax2.set_xlabel(r'$\nu$')
		ax2.set_ylabel('Target function')
		ax2.set_xscale('log')
		
		fig.tight_layout()        
		return x, fig, ax
	return x[0]


def guess_r(t, N, nu=None, K=None, PLOT=False):
	r"""Guesses the value of *r* from the shape of the growth curve.

	Following [Richards1959]_:

	.. math::

		\frac{dN}{dt}_{max} = r K \nu (1 + \nu)^{-\frac{1 + \nu}{\nu}}


	- :math:`\frac{dN}{dt}_{max}`: maximum population growth rate
	- r: initial per capita growth rate 
	- K: maximum population size
	- :math:`\nu`: curvature of the logsitic term

	Parameters
	----------
	t : numpy.ndarray
		time
	N : numpy:ndarray
		`N[i]` is the population size at time `t[i]`
	nu : float, optional
		a guess of `nu`, the maximum population size. If not given, it is guessed.
	K : float, optional
		a guess of `K`, the curvature of the logsitic term. If not given, it is guessed.

	Returns
	-------
	float
		the guess of *r*.

	References
	----------
	.. [Richards1959] Richards, F. J. 1959. `A Flexible Growth Function for Empirical Use <http://dx.doi.org/10.1093/jxb/10.2.290>`_. Journal of Experimental Botany
	"""
	t, N = np.array(t), np.array(N)
	if K is None:
		K = N.max()
	if nu is None:
		nu = guess_nu(t, N, K, PLOT=False, PRINT=False)
	
	# smooth and calculate derivative
	N_smooth = smooth(t, N) 	
	# calculate derivative	
	dNdt = derivative(N_smooth, t)
	# limit max search
	idx = N >= K * np.exp(-1) / 4
	t, dNdt = t[idx], dNdt[idx]
	i = dNdt.argmax()
	dNdtmax = dNdt[i]
	r = dNdtmax / (K * nu * (1 + nu)**(-(1 + nu) / nu))
	if PLOT:
		fs = plt.rcParams['figure.figsize']
		fig, ax = plt.subplots(1, 3, figsize=(fs[0] * 3, fs[1]))
		ax0, ax1, ax2 = ax
		
		N_t = N_smooth(t)
		ax0.plot(t, N_t, '-k')
		ax0.axvline(t[i], color='k', ls='--')
		ax0.axhline(N_t[i], color='k', ls='--')
		ax0.set_xlabel('Time')
		ax0.set_ylabel('N')
		
		ax1.plot(t, dNdt, '-k')
		ax1.axvline(t[i], color='k', ls='--')
		ax1.axhline(dNdt[i], color='k', ls='--')
		ax1.set_xlabel('Time')
		ax1.set_ylabel('dN/dt')
		dN = (dNdt.max() - dNdt.min()) / 10
		ax1.set_ylim(dNdt.min() - dN, dNdt.max() + dN)
		
		rs = dNdt / (K * nu * (1 + nu)**(-(1 + nu) / nu))
		ax2.plot(t, rs, '-k')
		ax2.axvline(t[i], color='k', ls='--')
		ax2.axhline(r, color='k', ls='--')
		ax2.set_xlabel('Time')
		ax2.set_ylabel('r')		

		fig.tight_layout()        
		return r, fig, ax
	return r


def guess_q0_v(t, N, param_guess, PLOT=False):
    r"""Guesses the values of :math:`q_0` and :math:`v` by fitting a model to a curve with other parameters fixed.
    
    Parameters
    ----------
    t : numpy.ndarray
        time
    N : numpy:ndarray
        `N[i]` is the population size at time `t[i]`
    param_guess : dict
        dictionary of parameter guesses for ``K``, ``y0``, ``nu``, ``r`` and possibly ``v``.
    PLOT : bool, optional
        if :const:`True`, the function will plot the calculations.
    
    Returns
    -------
    q0, v : float
        the guess of :math:`\q_0` and ``v``.
    fig : matplotlib.figure.Figure
        if the argument `PLOT` was :const:`True`, the generated figure.
    ax : matplotlib.axes.Axes
        if the argument `PLOT` was :const:`True`, the generated axis.
    """
    param_guess = dict(param_guess)
    param_fix = set(param_guess.keys())    
    param_guess['q0'] = param_guess.get('q0', 1) # default guess is 1
    param_guess['v'] = param_guess.get('v', 1) # default guess is 1
    model = curveball.baranyi_roberts_model.BaranyiRoberts()
    params = model.guess(data=N, t=t, param_guess=param_guess, param_fix=param_fix)
    result = model.fit(data=N, t=t, params=params)
    if PLOT:
        ax = result.plot_fit()
        return result.best_values['q0'], result.best_values['v'], ax.figure, ax
    return result.best_values['q0'], result.best_values['v']


class BaranyiRoberts(lmfit.model.Model):
	"""The Baranyi-Roberts growth model is an extension of the logistic and Richards model that adds a lag phase.	
	"""

	def __init__(self, *args, **kwargs):
		if args:
			func, args = args[0], args[1:]
		else:
			func = baranyi_roberts_function
		super(BaranyiRoberts, self).__init__(func, *args, **kwargs)
		self.name = self.__class__.__name__
		self.nested_models = {
			'nu': LogisticLag2,
			'lag': Richards
		}


	def guess(self, data, t, param_guess=None, param_min=None, param_max=None, param_fix=None):
		"""Use heuristics to guess model parameters.

		Parameters
		----------
		data : numpy.ndarray
			population size or density data
		t : numpy.ndarray
			time, usually in hours
		param_guess : dict, optional
			mapping parameter names to values used as guesses to override the heuristic guess
		param_min, param_max : dict, optional
			mapping parameter names to values used to bound the parametrs in the fit procedure
		param_fix : set, optional
			names of parameters to fix at the guessed value, i.e., no to estimate in the fit procedure

		Returns
		-------
		lmfit.parameter.Parameters
			a dictionary of the model parameters ready to be used by the `fit` method

		See also
		--------
		lmfit.model.Model.guess
		"""	
		if (sorted(t) != t).all():
			raise ValueError("Time argument t must be sorted")

		param_guess = dict() if param_guess is None else dict(param_guess)
		param_min = dict() if param_min is None else dict(param_min)
		param_max = dict() if param_max is None else dict(param_max)
		param_fix = set() if param_fix is None else set(param_fix)
		
		if 'y0' not in param_guess:
			param_guess['y0'] = (data[t == t.min()]).mean()
		if 'K' not in param_guess:
			param_guess['K'] = (data[t == t.max()]).mean()
		if 'nu' not in self.param_names:
			nu = 1.0
		elif 'nu' not in param_guess:
			# guess_nu has not been performing well, just use 1.0
			# param_guess['nu'] = guess_nu(t, data, K=param_guess['K'], PLOT=False, PRINT=False)
			param_guess['nu'] = 1.0 
			nu = param_guess['nu']
		else:
			nu = param_guess['nu']
		if 'r' not in param_guess:
			param_guess['r'] = guess_r(t, data, K=param_guess['K'], nu=nu)		
		if 'q0' not in param_guess:
			if 'v' not in self.param_names or 'v' in param_guess:
				param_guess['q0'], _ = guess_q0_v(t, data, param_guess)
			else:
				param_guess['q0'], param_guess['v'] = guess_q0_v(t, data, param_guess)

		params = lmfit.parameter.Parameters()
		for pname in self.param_names:		
			params.add(
				name 	= pname,
				value 	= param_guess[pname],
				min 	= param_min.get(pname, MIN_VALUES.get(pname, 0)),
				max 	= param_max.get(pname, MAX_VALUES.get(pname, np.inf)),
				vary 	= pname not in param_fix
			)
		# if 'q0' in params:
		# 	if 'v' in params:
		# 		params.add('lag', value=_lag(q0=param_guess['q0'], v=param_guess['v']), min=1.0/60.0, vary=params['q0'].vary)
		# 		params['q0'].set(expr='1/(exp(v / 60.0) - 1)')
		# 	else:
		# 		params.add('lag', value=_lag(q0=param_guess['q0'], v=param_guess['r']), min=1.0/60.0, vary=params['q0'].vary)
		# 		params['q0'].set(expr='1/(exp(r / 60.0) - 1)')
			
		return params


	def get_sympy_expr(self, params):
		"""Generate the required variables for creating a Dfun to be used in the fitting procedure.

		Parameters
		----------
		params : dict
			a dictionary of the model parameters, possibly created by :py:method:`guess`.

		Returns
		-------
			dNdt : sympy.expr.Expr
				expression for the derivate of the model function
			t : sympy.symbol.symbol
			ma	time symbol
			args : tuple of sympy.symbol.symbol
				tuple of symbols of model parameters

		See also
		--------
		curveball.models.make_Dfun
		"""
		t, y0, K, r, nu, q0, v = sympy.symbols('t y0 K r nu q0 v')
		args = [y0, K, r, nu, q0, v]
		# remove fixed params and replace their symbols by their values
		# for params that don't exist (in inherited models), remove them and replace with a default value
		if not params['y0'].vary:
			args.remove(y0)
			y0 = params['y0'].value
		if 'K' not in self.param_names:
			args.remove(K)
			K = np.inf
		elif not params['K'].vary:
			args.remove(K)
			K = params['K'].value
		if not params['r'].vary:
			args.remove(r)
			r = params['r'].value
		if 'nu' not in self.param_names:
			args.remove(nu)
			nu = 1.0
		elif not params['nu'].vary:
			args.remove(nu)
			nu = params['nu'].value
		if 'q0' not in self.param_names:
			args.remove(q0)
			q0 = np.inf
		elif not params['q0'].vary:
			args.remove(q0)
			q0 = params['q0'].value
		if 'v' not in self.param_names:
			args.remove(v)
			v = r
		elif not params['v'].vary:
			args.remove(v)
			v = params['v'].value
		if (isinstance(q0, float) and np.isinf(q0)) or (isinstance(v, float) and np.isinf(v)):
			At = t
		else:
			At = t + 1.0 / v * sympy.log((sympy.exp(-v * t) + q0) / (1 + q0))
		dNdt = K / (1.0 - (1.0 - (K / y0)**nu) * sympy.exp(-r * nu * At))**(1.0 / nu)
		return dNdt, t, tuple(args)


class Richards(BaranyiRoberts):
	r"""The Richards model [Richards1959]_ extends the logistic model by including the :math:`nu` parameter that controls the
	curvature of the response of the growth rate to population size/density. Also knows as the :math:`\theta`-logistic model[Gilpin1973]_.

	References
	----------
	.. [Richards1959] Richards, F. J. 1959. "A Flexible Growth Function for Empirical Use." Journal of Experimental Botany 10 (2): 290-301. doi:10.1093/jxb/10.2.290.
	.. [Gilpin1973] Gilpin, Michael E., and Francisco J. Ayala. 1973. "Global Models of Growth and Competition." Proceedings of the National Academy of Sciences of the United States of America 70 (12 Pt 1-2): 3590–3593. doi:10.1073/pnas.70.12.3590.
	"""
	def __init__(self, *args, **kwargs):
		def func(t, y0, K, r, nu):			
			return baranyi_roberts_function(t, y0, K, r, nu, np.inf, np.inf)
		super(Richards, self).__init__(func, *args, **kwargs)
		self.nested_models = {
			'nu': Logistic
		}

class RichardsLag1(BaranyiRoberts):
	r"""This is an extension of the :py:class:`Richards` that includes a single lag parameter :math:`q_0` (and sets :math:`v=r`).
	"""
	def __init__(self, *args, **kwargs):
		def func(t, y0, K, r, nu, q0): 
			return baranyi_roberts_function(t, y0, K, r, nu, q0, r)
		super(RichardsLag1, self).__init__(func, *args, **kwargs)
		self.nested_models = {
			'nu': LogisticLag1,
			'lag': Richards
		}

class LogisticLag2(BaranyiRoberts):
	r"""This is an extension of the :py:class:`Logistic` model that includes a two lag parameters :math:`q_0` and :math:`v`).	
	"""
	def __init__(self, *args, **kwargs):
		def func(t, y0, K, r, q0, v): 
			return baranyi_roberts_function(t, y0, K, r, 1.0, q0, v)
		super(LogisticLag2, self).__init__(func, *args, **kwargs)
		self.nested_models = {
			'lag': Logistic
		}

class LogisticLag1(BaranyiRoberts):
	r"""This is an extension of the :py:class:`Logistic` that includes a single lag parameter :math:`q_0` (and sets :math:`v=r`).
	"""
	def __init__(self, *args, **kwargs):
		def func(t, y0, K, r, q0):
			return baranyi_roberts_function(t, y0, K, r, 1.0, q0, r)
		super(LogisticLag1, self).__init__(func, *args, **kwargs)
		self.nested_models = {
			'lag': Logistic
		}

class Logistic(BaranyiRoberts):
	r"""This is the simplest growth model with a stationary phase, defined by the ODE:

	.. math::
	
		\frac{dN}{dt} = r N \Big( 1 - \frac{N}{K} \Big)
	
	Notes
	-----
	`Wikipedia <https://en.wikipedia.org/wiki/Logistic_function#In_ecology:_modeling_population_growth>`_
	"""
	def __init__(self, *args, **kwargs):
		def func(t, y0, K, r):
			return baranyi_roberts_function(t, y0, K, r, 1.0, np.inf, np.inf)
		super(Logistic, self).__init__(func, *args, **kwargs)
		self.nested_models = {}


if __name__ == '__main__':
	def nvarys(params): 
		return len([p for p in params.values() if p.vary])
	from curveball.models import randomize

	t, y = randomize(t=12, y0=0.12, K=0.56, r=0.8, nu=3.0, q0=0.2, v=0.8, reps=1, as_df=False, random_seed=0)
	plt.plot(t, y, 'o')
	plt.show()
	print("Step", USE_STEP_FUNC)

	br6_model = BaranyiRoberts()
	br6_params = br6_model.guess(data=y, t=t)
	assert nvarys(br6_params) == 6
	br6_result = br6_model.fit(data=y, t=t, params=br6_params)
	assert br6_result.nvarys == 6
	print(br6_result.model.name, br6_result.nvarys, br6_result.bic, _lag(br6_result))

	br5_model = LogisticLag2()
	br5_params = br5_model.guess(data=y, t=t, param_guess={'nu':1}, param_fix=['nu'])
	assert nvarys(br5_params) == 5, br5_params
	assert br5_params['nu'].value == 1
	br5_result = br5_model.fit(data=y, t=t, params=br5_params)
	assert br5_result.nvarys == 5
	assert br5_result.best_values['nu'] == 1
	print(br5_result.model.name, br5_result.nvarys, br5_result.bic, _lag(br5_result))


	br5b_model = RichardsLag1()
	br5b_params = br5b_model.guess(data=y, t=t, param_fix=['v'])
	assert nvarys(br5b_params) == 5, br5b_params
	assert br5b_params['v'].value == br5b_params['r'].value
	br5b_result = br5b_model.fit(data=y, t=t, params=br5b_params)
	assert br5b_result.nvarys == 5
	assert br5b_result.best_values['v'] == br5b_result.best_values['r']
	print(br5b_result.model.name, br5b_result.nvarys, br5b_result.bic, _lag(br5b_result))


	br4_model = LogisticLag1()
	br4_params = br4_model.guess(data=y, t=t, param_guess={'nu':1}, param_fix=['nu', 'v'])
	assert nvarys(br4_params) == 4, br4_params
	assert br4_params['nu'].value == 1
	assert br4_params['v'].value == br4_params['r'].value
	br4_result = br4_model.fit(data=y, t=t, params=br4_params)
	assert br4_result.nvarys == 4
	assert br4_result.best_values['v'] == br4_result.best_values['r']
	print(br4_result.model.name, br4_result.nvarys, br4_result.bic, _lag(br4_result))

	richards_model = Richards()
	richards_params = richards_model.guess(data=y, t=t, param_guess={'v':np.inf}, param_fix=['q0', 'v'])
	assert nvarys(richards_params) == 4
	assert np.isposinf(richards_params['v'].value)
	assert np.isposinf(richards_params['q0'].value)
	richards_result = richards_model.fit(data=y, t=t, params=richards_params)
	assert richards_result.nvarys == 4
	assert np.isposinf(richards_result.best_values['v'])
	assert np.isposinf(richards_result.best_values['q0'])
	print(richards_result.model.name, richards_result.nvarys, richards_result.bic, _lag(richards_result))

	logistic_model = Logistic()
	logistic_params = logistic_model.guess(data=y, t=t, param_guess={'nu':1, 'v':np.inf}, param_fix=['nu', 'v', 'q0'])
	assert nvarys(logistic_params) == 3
	assert logistic_params['nu'].value == 1
	assert np.isposinf(logistic_params['v'].value)
	assert np.isposinf(logistic_params['q0'].value)
	logistic_result = logistic_model.fit(data=y, t=t, params=logistic_params)
	assert logistic_result.nvarys == 3
	assert logistic_result.best_values['nu'] == 1
	assert np.isposinf(logistic_result.best_values['v'])
	assert np.isposinf(logistic_result.best_values['q0'])
	print(logistic_result.model.name, logistic_result.nvarys, logistic_result.bic, _lag(logistic_result))



from builtins import map
#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
from unittest import TestCase, main
import pytest
import os
import glob
import io
import shutil
from importlib import resources
import pandas as pd
from click.testing import CliRunner # ...
from curveball.scripts import cli
import curveball
import click
import numpy as np


CI = os.environ.get('CI', 'false').lower() == 'true'


def is_csv(data):
	lines = data.splitlines()
	data  = [line.split(',') for line in lines]
	lengths = list(map(len, data))
	return all(x==lengths[0] for x in lengths)


class SimpleTestCase(TestCase):
	_multiprocess_can_split_ = True

	def setUp(self):
		self.runner = CliRunner()


	def tearDown(self):
		pass


	def test_help(self):
		result = self.runner.invoke(cli.cli, ['--help'])
		self.assertEqual(result.exit_code, 0)


	def test_version(self):
		result = self.runner.invoke(cli.cli, ['--version'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn(curveball.__version__, result.output)


	def test_where(self):
		result = self.runner.invoke(cli.cli, ['--where'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn("curveball", result.output.lower())
		path = result.output.strip()
		self.assertTrue(os.path.exists(path), msg=path)


class PlateTestCase(TestCase):
	_multiprocess_can_split_ = True


	def setUp(self):
		self.runner = CliRunner()


	def _is_plate_csv(self, data):
		self.assertTrue(is_csv(data))
		newlines = data.count("\n")
		self.assertEqual(newlines, 97) # 96 wells, 1 header line


	def test_default_plate(self):
		result = self.runner.invoke(cli.cli, ['plate'])
		self.assertEqual(result.exit_code, 0)
		self._is_plate_csv(result.output)


	def test_non_default_plate(self):
		result = self.runner.invoke(cli.cli, ['plate', '--plate_file=G-RG-R.csv'])
		self.assertEqual(result.exit_code, 0)
		self._is_plate_csv(result.output)


	def test_default_plate_to_file(self):
		filename = 'plate.csv'
		with self.runner.isolated_filesystem():
			result = self.runner.invoke(cli.cli, ['plate', '--output_file={0}'.format(filename)])
			self.assertEqual(result.exit_code, 0)
			with open(filename, 'r') as f:
				data = f.read()		
			self._is_plate_csv(data)


	def test_plate_not_found(self):
		result = self.runner.invoke(cli.cli, ['plate', '--plate_file=untitled.csv'])
		self.assertNotEqual(result.exit_code, 0)
		self.assertIn('untitled.csv', result.output)


	def test_bad_plate_file(self):
		filename = 'bad_plate.csv'
		with self.runner.isolated_filesystem():
			with open(filename, 'w') as f:
				import this
				f.write(this.s)
			result = self.runner.invoke(cli.cli, ['plate', '--plate_file={0}'.format(filename)])
			self.assertNotEqual(result.exit_code, 0, result.output)
			self.assertIn(filename, result.output)


	def test_plate_list(self):
		result = self.runner.invoke(cli.cli, ['plate', '--list'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('G-RG-R.csv', result.output)
		self.assertTrue(result.output.count('\n') > 2)


	def test_plate_plot(self):
		result = self.runner.invoke(cli.cli, ['plate', '--show'])
		self.assertEqual(result.exit_code, 0)		


	def test_plate_plot_to_file(self):
		filename = 'plate.png'
		with self.runner.isolated_filesystem():
			result = self.runner.invoke(cli.cli, ['plate', '--show', '--output_file={0}'.format(filename)])
			self.assertEqual(result.exit_code, 0)
			self.assertTrue(os.path.exists(filename))		


class AnalysisTestCase(TestCase):
	_multiprocess_can_split_ = True


	def setup_with_context_manager(self, cm):
	    """Use a contextmanager to setUp a test case.
	    See http://nedbatchelder.com/blog/201508/using_context_managers_in_test_setup.html
	    """
	    val = cm.__enter__()
	    self.addCleanup(cm.__exit__, None, None, None)
	    return 


	def setUp(self):
		self.files = [p.name for p in resources.files('data').iterdir()]
		self.files = [fn for fn in self.files if os.path.splitext(fn)[-1] in ['.xlsx', '.mat']]
		self.files = [fn for fn in self.files if not fn.lower().startswith('sunrise')]
		self.files = [fn for fn in self.files if not fn.lower().startswith('biotek')]
		self.files.sort()
		self.runner = CliRunner()
		self.ctx = self.setup_with_context_manager(self.runner.isolated_filesystem())
		self.dirpath = os.getcwd()
		self.assertTrue(os.path.exists(self.dirpath))
		self.assertTrue(os.path.isdir(self.dirpath))

		print("files:", self.files)
		for fn in self.files:
			src = resources.files('data').joinpath(fn)
			shutil.copy(src, '.')
			self.assertTrue(os.path.exists(os.path.join(self.dirpath, fn)))
			self.assertTrue(os.path.isfile(os.path.join(self.dirpath, fn)))
		self.filepath = os.path.join(self.dirpath, self.files[1])
		

	def tearDown(self):
		pass	


	def test_process_file(self):
		result = self.runner.invoke(cli.cli, ['--no-plot', '--verbose', '--no-prompt', 'analyse', self.filepath, '--plate_file=G-RG-R.csv', '--ref_strain=G'])
		self.assertEqual(result.exit_code, 0, "Code: {}\n{}".format(result.exit_code, result.output))
		lines = [line for line in result.output.splitlines() if len(line) > 0] 
		data = os.linesep.join(lines[-4:]) # only last 4 lines
		self.assertTrue(is_csv(data))


	def test_process_file2(self):
		filepath = os.path.join(self.dirpath, self.files[1])
		result = self.runner.invoke(cli.cli, ['--no-plot', '--verbose', '--no-prompt', 'analyse', filepath, '--plate_file=G-RG-R.csv', '--ref_strain=G'])
		self.assertEqual(result.exit_code, 0, "Code: {}\n{}".format(result.exit_code, result.output))
		lines = [line for line in result.output.splitlines() if len(line) > 0] 
		data = os.linesep.join(lines[-4:]) # only last 4 lines
		self.assertTrue(is_csv(data))


	def test_process_file_to_file(self):
		result = self.runner.invoke(cli.cli, ['--no-plot', '--verbose', '--no-prompt', 'analyse', self.filepath, '--plate_file=G-RG-R.csv', '--ref_strain=G', '--output_file=summary.csv'])
		self.assertEqual(result.exit_code, 0, "Code: {}\n{}".format(result.exit_code, result.output))
		with open('summary.csv', 'r') as f:
			data = f.read()
		self.assertTrue(is_csv(data))


	def test_process_file_with_guess(self):
		result = self.runner.invoke(cli.cli, ['--no-plot', '--verbose', '--no-prompt', 'analyse', self.filepath, '--guess', 'K', '0.7', '--guess', 'nu', '2.0', '--plate_file=G-RG-R.csv', '--ref_strain=G'])
		self.assertEqual(result.exit_code, 0, "Code: {}\n{}".format(result.exit_code, result.output))
		lines = [line for line in result.output.splitlines() if len(line) > 0] 
		data = os.linesep.join(lines[-4:]) # only last 4 lines
		self.assertTrue(is_csv(data))


	def test_process_file_with_param_max(self):
		result = self.runner.invoke(cli.cli, ['--no-plot', '--verbose', '--no-prompt', 'analyse', self.filepath, '--param_max', 'K', '0.7', '--param_max', 'nu', '2.0', '--param_max', 'v', '100', '--plate_file=G-RG-R.csv', '--ref_strain=G'])
		self.assertEqual(result.exit_code, 0, "Code: {}\n{}".format(result.exit_code, result.output))
		lines = [line for line in result.output.splitlines() if len(line) > 0] 
		data = os.linesep.join(lines[-4:]) # only last 4 lines
		self.assertTrue(is_csv(data))


	def test_process_file_with_param_min(self):
		result = self.runner.invoke(cli.cli, ['--no-plot', '--verbose', '--no-prompt', 'analyse', self.filepath, '--param_min', 'K', '0.2', '--param_min', 'nu', '0.1', '--param_min', 'v', '0.1', '--plate_file=G-RG-R.csv', '--ref_strain=G'])
		self.assertEqual(result.exit_code, 0, "Code: {}\n{}".format(result.exit_code, result.output))
		lines = [line for line in result.output.splitlines() if len(line) > 0] 
		data = os.linesep.join(lines[-4:]) # only last 4 lines
		self.assertTrue(is_csv(data))


	def test_process_file_with_param_fix(self):
		result = self.runner.invoke(cli.cli, ['--no-plot', '--verbose', '--no-prompt', 'analyse', self.filepath, '--param_fix', 'K', '--plate_file=G-RG-R.csv', '--ref_strain=G'])
		#self.assertEqual(result.exit_code, 0, "Code: {}\n{}".format(result.exit_code, result.output))
		lines = [line for line in result.output.splitlines() if len(line) > 0] 
		data = os.linesep.join(lines[-4:]) # only last 4 lines
		self.assertTrue(is_csv(data))


	def test_process_file_without_weights(self):
		result = self.runner.invoke(cli.cli, ['--no-plot', '--verbose', '--no-prompt', 'analyse', self.filepath, '--no-weights', '--plate_file=G-RG-R.csv', '--ref_strain=G'])
		self.assertEqual(result.exit_code, 0, "Code: {}\n{}".format(result.exit_code, result.output))
		lines = [line for line in result.output.splitlines() if len(line) > 0] 
		data = os.linesep.join(lines[-4:]) # only last 4 lines
		self.assertTrue(is_csv(data))

	def test_constant_od_emits_nan(self):
		filename = 'constant.csv'
		with self.runner.isolated_filesystem():
			constant_df = pd.DataFrame({
				'Time': [0, 1],
				'OD': [0.1, 0.1],
				'Well': ['A1', 'A1'],
				'Strain': ['G', 'G'],
				'Row': ['A', 'A'],
				'Col': [1, 1]
			})
			constant_df.to_csv(filename, index=False)
			result = self.runner.invoke(
				cli.cli, ['--no-plot', '--verbose', '--no-prompt', 'analyse', filename,
					'--output_file=out.csv', '--plate_file=G-RG-R.csv', '--ref_strain=G']
			)
			self.assertEqual(result.exit_code, 0, result.output)
			output_df = pd.read_csv('out.csv')
			self.assertIn('NRMSD', output_df.columns)
			self.assertTrue(np.isnan(output_df.loc[0, 'NRMSD']))
			self.assertTrue(np.isnan(output_df.loc[0, 'CV(RMSD)']))


	# this test works but takes too long (10 min) see #129
	def test_process_file_with_ci(self):
		result = self.runner.invoke(cli.cli, [
			'--no-plot', '--verbose', '--no-prompt',
			'analyse', self.filepath, '--ci', '--nsamples=3',
			'--plate_file=G-RG-R.csv', '--ref_strain=G'
		])
		self.assertEqual(result.exit_code, 0, "Code: {}\n{}".format(result.exit_code, result.output))
		lines = [line for line in result.output.splitlines() if len(line) > 0]
		data = os.linesep.join(lines[-4:]) # only last 4 lines
		self.assertTrue(is_csv(data))


	# FIXME - fails on CI, succeeds on local
	# def test_create_plots(self):
	# 	result = self.runner.invoke(cli.cli, ['--plot', '--verbose', '--no-prompt', 'analyse', self.filepath, '--plate_file=G-RG-R.csv', '--ref_strain=G'])
	# 	self.assertEqual(result.exit_code, 0, result.output)
	# 	path,ext = os.path.splitext(self.filepath)		
	# 	plot_files = glob.glob(path + "_*.png")
	# 	self.assertNotEqual(len(plot_files), 0, result.output)


	def test_process_folder(self):
		num_files = len(self.files)
		result = self.runner.invoke(cli.cli, ['--no-plot', '--verbose', '--no-prompt', 
			'analyse', self.dirpath, '--plate_file=G-RG-R.csv', '--ref_strain=G'])
		self.assertEqual(result.exit_code, 0, "Code: {}\n{}".format(result.exit_code, result.output))
		lines = [line for line in result.output.splitlines() if len(line) > 0] 
		num_lines =  num_files * 3 + 1
		data = os.linesep.join(lines[-num_lines:])
		self.assertTrue(is_csv(data), result.output)
		

	def test_path_not_found(self):
		result = self.runner.invoke(cli.cli, ['analyse', 'untitled.xlsx'])
		self.assertNotEqual(result.exit_code, 0)
		self.assertIn('untitled.xlsx', result.output)


	def test_no_files_in_folder(self):
		for fn in self.files:
			os.remove(fn)
		self.assertEqual(len(glob.glob("*")), 0)
		result = self.runner.invoke(cli.cli, ['--no-plot', '--verbose', '--no-prompt', 'analyse', '.'])
		self.assertNotEqual(result.exit_code, 0)
		self.assertIn('.', result.output)


	def test_bad_data_file(self):
		shutil.copyfile(__file__, 'untitled.xlsx')
		result = self.runner.invoke(cli.cli, ['analyse', 'untitled.xlsx'])
		self.assertNotEqual(result.exit_code, 0)
		self.assertIn('untitled.xlsx', result.output)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
from unittest import TestCase, main
from builtins import str
import tempfile
import os
import shutil
import zipfile
from importlib import resources
import curveball
import pandas as pd
from pandas.testing import assert_frame_equal


RANDOM_SEED = int(os.environ.get('RANDOM_SEED', 0))


class CurveballCSVTestCase(TestCase):
    def setUp(self):
        self.filename = resources.files("data").joinpath("Tecan_210115.csv")
        self.plate = pd.read_csv(resources.files("plate_templates").joinpath("G-RG-R.csv"))

    def test_read_curveball_csv(self):
        df = curveball.ioutils.read_curveball_csv(self.filename)
        self.assertIsNotNone(df)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(df.shape, (4992, 9))
        self.assertEqual(df.columns.tolist() , ['Time', u'Temp. [\xb0C]', 'Cycle Nr.', 'Well', 'OD', 'Row', 'Col', 'Strain', 'Color'])

    def test_write_curveball_csv(self):
        df = curveball.models.randomize(random_seed=RANDOM_SEED)
        f, fn = tempfile.mkstemp()
        curveball.ioutils.write_curveball_csv(df, fn)
        df1 = pd.read_csv(fn)
        assert_frame_equal(df, df1, check_dtype=False)


class TecanXLSXTestCase(TestCase):
    def setUp(self):
        self.filename = resources.files("data").joinpath("Tecan_210115.xlsx")
        self.plate = pd.read_csv(resources.files("plate_templates").joinpath("G-RG-R.csv"))

    def test_read_tecan_xlsx_OD(self):
        df = curveball.ioutils.read_tecan_xlsx(self.filename, 'OD')
        self.assertIsNotNone(df)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(df.shape, (8544, 9))
        self.assertEqual(df.columns.tolist() , ['Time', u'Temp. [\xb0C]', 'Cycle Nr.', 'Well', 'OD', 'Row', 'Col', 'Strain', 'Color'])

    def test_read_tecan_xlsx_full(self):
        df = curveball.ioutils.read_tecan_xlsx(self.filename, ('OD', 'Green', 'Red'))
        self.assertIsNotNone(df)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(df.shape, (8352, 15))
        self.assertEqual(sorted(df.columns.tolist()) , sorted(['Time_OD', u'Temp. [\xb0C]_OD', 'Cycle Nr.', 'Well', 'OD', 'Row', 'Col', 'Strain', 'Color', 'Time_Green', u'Temp. [\xb0C]_Green', 'Green', 'Time', u'Temp. [\xb0C]', 'Red']))

    def test_read_tecan_xlsx_12hrs(self):
        df = curveball.ioutils.read_tecan_xlsx(self.filename, 'OD', max_time=12)
        self.assertIsNotNone(df)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(df.shape, (4992, 9))
        self.assertEqual(df.columns.tolist() , ['Time', u'Temp. [\xb0C]', 'Cycle Nr.', 'Well', 'OD', 'Row', 'Col', 'Strain', 'Color'])

    def test_read_tecan_xlsx_with_plate(self):
        df = curveball.ioutils.read_tecan_xlsx(self.filename, 'OD', plate=self.plate)
        self.assertIsNotNone(df)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(df.shape, (8544, 9))
        self.assertEqual(df.columns.tolist() , ['Time', u'Temp. [\xb0C]', 'Cycle Nr.', 'Well', 'OD', 'Row', 'Col', 'Strain', 'Color'])


class BioTekXLSXTestCase(TestCase):
    def setUp(self):
        self.filename = resources.files("data").joinpath("BioTekSynergy.xlsx")

    def test_read_biotek_xlsx(self):
        df = curveball.ioutils.read_biotek_xlsx(self.filename)
        self.assertIsNotNone(df)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(df.columns.tolist(), ['Time', u'T\xb0 600', 'Well', 'OD', 'Row', 'Col', 'Strain', 'Color'])


# class XMLTestCase(TestCase):
#     def setUp(self):        
#         self.zip_filename = pkg_resources.resource_filename("data", "Tecan_210115.xlsx")
#         self.zip_filename = os.path.join("data", "20130211_dh.zip")
#         self.folder = tempfile.mkdtemp()
#         self.zipfile = zipfile.ZipFile(self.zip_filename)
#         self.zipfile.extractall(self.folder)
#         self.plate = pd.read_csv(pkg_resources.resource_filename("plate_templates", "checkerboard.csv"))


#     def tearDown(self):
#         shutil.rmtree(self.folder)


#     def test_read_tecan_xml_with_plate(self):
#         df = curveball.ioutils.read_tecan_xml(os.path.join(self.folder, "*.xml"), 'OD', plate=self.plate)
#         self.assertIsNotNone(df)
#         self.assertIsInstance(df, pd.DataFrame)
#         self.assertEqual(df.shape, (2016, 8))
#         self.assertEqual(df.columns.tolist() , ['OD', 'Well', 'Row', 'Col', 'Time', 'Filename', 'Strain', 'Color'])    


class SunriseTestCase(TestCase):
    def setUp(self):
        self.filename = resources.files("data").joinpath("Sunrise_180515_0916.xlsx")
        self.plate = pd.read_csv(resources.files("plate_templates").joinpath("G-RG-R.csv"))

    def test_read_sunrise_xlsx(self):
        df = curveball.ioutils.read_sunrise_xlsx(self.filename)
        self.assertIsNotNone(df)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(df.shape, (96, 8))
        self.assertEqual(sorted(df.columns.tolist()) , sorted([u'Time', u'Well', u'OD', u'Row', u'Col', 'Strain', 'Color', 'Filename']))

    # commented out 18JUN19 due to travis-ci failure:
    # ValueError: You are trying to merge on object and int64 columns. If you wish to proceed you should use pd.concat
    # def test_read_sunrise_xlsx_plate(self):
    #     df = curveball.ioutils.read_sunrise_xlsx(self.filename, plate=self.plate)
    #     self.assertIsNotNone(df)
    #     self.assertIsInstance(df, pd.DataFrame)
    #     self.assertEqual(df.shape, (96, 8))
    #     self.assertEqual(sorted(df.columns.tolist()) , sorted([u'Time', u'Well', u'OD', u'Row', u'Col', 'Strain', 'Color', 'Filename']))


class MatTestCase(TestCase):
    def setUp(self):
        self.filename = resources.files("data").joinpath("plate_9_OD.mat")
        self.plate = pd.read_csv(resources.files("plate_templates").joinpath("checkerboard.csv"))

    def test_read_tecan_mat(self):
        df = curveball.ioutils.read_tecan_mat(self.filename, plate=self.plate)
        self.assertIsNotNone(df)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(df.shape, (2496, 8))
        self.assertEqual(df.columns.tolist() , [u'Cycle Nr.', u'Time', u'Well', u'OD', u'Row', u'Col', 'Strain', 'Color'])


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>import pandas as pd
import curveball

if __name__ == '__main__':
	plate = pd.read_csv('plate_templates/G-RG-R.csv')
	df = curveball.ioutils.read_tecan_xlsx('data/Tecan_280715.xlsx', label='OD', plate=plate)
	green = curveball.models.fit_model(df[df.Strain == 'G'], PLOT=False, PRINT=False)[0]
	red = curveball.models.fit_model(df[df.Strain == 'R'], PLOT=False, PRINT=False)[0]
	t, y, fig, ax = curveball.competitions.compete(green, red, PLOT=True, colors=['green', 'red'])
	fig.savefig('docs/_static/example_competition.svg')


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>import pandas as pd
import curveball

if __name__ == '__main__':
	plate = pd.read_csv('plate_templates/G-RG-R.csv')
	fig, ax = curveball.plots.plot_plate(plate)	
	fig.savefig('docs/_static/example_plot_plate.svg')


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
from __future__ import print_function
from __future__ import division
from builtins import str
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
sns.set_style("ticks")
import curveball
import curveball.baranyi_roberts_model


def loglik(t, y, y_sig, f, penalty=None, **params):
    r"""Computes the log-likelihood of seeing the data given a model 
    assuming normal distributed observation/measurement errors.

    .. math::

        \log{L(y | \theta)} = -\frac{1}{2} \sum_i { \log{(2 \pi \sigma_{i}^{2})} + \frac{(y - f(t_i; \theta))^2}{\sigma_{i}^{2}} }

    which is the log-likelihood of seeing the data points :math:`t_i, y_i` 
    with measurement error :math:`\sigma_i`
    given the model function :math:`f`, the model parameters :math:`\theta`, 
    and that the measurement error at time :math:`t_i` has a normal distribution with mean 0.

    Parameters
    ----------
    t : np.ndarray
        one dimensional array of time
    y : np.ndarray
        one dimensional array of the means of the observations 
    y_sig : np.ndarray
        one dimensional array of standrad deviations of the observations
    f : callable
        a function the calculates the expected observations (`f(t)`) from `t` and any parameters in `params`
    penalty : callable
        a function that calculates a scalar penalty from the parameters in `params` to be substracted from the log-likelihood
    params : floats, optional
        model parameters

    Returns
    -------
    float
        the log-likelihood result
    """
    yhat = f(t, **params)
    val = -0.5 * np.sum(np.log(2 * np.pi * y_sig ** 2) + (y - yhat) ** 2 / y_sig ** 2)
    if penalty:
        val -= penalty(**params)
    return val
        

def ridge_regularization(lam, **center):
    r"""Create a penaly function that employs the ridge regularization method:

    .. math::

        P = \lambda ||\theta - \theta_0||_2

    where :math:`\lambda` is the regularization scale, 
    :math:`\theta` is the model parameters vector, 
    and :math:`\theta_0` is the model parameters guess vector.
    This is similar to using a multivariate Gaussian prior distribution on the model parameters 
    with the Gaussian centerd at :math:`\theta_0` and scaled by :math:`\lambda`.

    Parameters
    ----------
    lam : float
        the penalty factor or regularization scale
    center : floats, optional
        guesses of model parameters

    Returns
    -------
    callable
        the penalty function, accepts model parameters as float keyword arguments and returns a float penalty to the log-likelihood

    Examples
    --------
    >>> penalty = ridge_regularization(1, y=0.1, K=1, r=1)
    >>> loglik(t, y, y_sig, logistic, penalty=penalty, y0=0.12, K=0.98, r=1.1)
    """
    def _ridge_regularization(**params):
        return lam * np.linalg.norm([v - center.get(k, 0) for k, v in params.items() if np.isfinite(v)])
    return _ridge_regularization


def loglik_r_nu(r_range, nu_range, df, f=curveball.baranyi_roberts_model.baranyi_roberts_function, 
                penalty=None, **params):
    r"""Estimates the log-likelihood surface for :math:`r` and :math:`\nu` given data and a model function.

    Parameters
    ----------
    r_range, nu_range : numpy.ndarray
        vectors of floats of :math:`r` and :math:`\nu` values on which to compute the log-likelihood
    df : pandas.DataFrame
        data frame with `Time` and `OD` columns
    f : callable, optional
        model function, defaults to :py:func:`curveball.baranyi_roberts_model.baranyi_roberts_function`
    penalty : callable, optional
        if given, the result of `penalty` will be substracted from the log-likelihood for each parameter set
    params : floats
        values for the model model parameters used by `f`

    Returns
    -------
    np.ndarray
        two-dimensional array of log-likelihood calculations;
        value at index `i, j` will have the value for `r_range[i]` and `nu_range[j]`

    See also
    --------
    loglik
    loglik_r_q0
    """
    if not params:
        params = dict()
    t = df.Time.unique()
    y = df.groupby('Time')['OD'].mean().to_numpy()
    y_sig = df.groupby('Time')['OD'].std().to_numpy()
    
    output = np.empty((len(r_range), len(nu_range)))
    for i, r in enumerate(r_range):
        params['r'] = r
        for j, nu in enumerate(nu_range):
            params['nu'] = nu
            output[i,j] = loglik(t, y, y_sig, f, penalty, **params)
    
    return output


def loglik_r_q0(r_range, q0_range, df, f=curveball.baranyi_roberts_model.baranyi_roberts_function, 
                penalty=None, **params):
    r"""Estimates the log-likelihood surface for :math:`r` and :math:`\nu` given data and a model function.

    Parameters
    ----------
    r_range, q0_range : numpy.ndarray
        vectors of floats of :math:`r` and :math:`q_0` values on which to compute the log-likelihood
    df : pandas.DataFrame
        data frame with `Time` and `OD` columns
    f : callable, optional
        model function, defaults to :py:func:`curveball.baranyi_roberts_model.baranyi_roberts_function`
    penalty : callable, optional
        if given, the result of `penalty` will be substracted from the log-likelihood for each parameter set
    params : floats
        values for the model model parameters used by `f`

    Returns
    -------
    np.ndarray
        two-dimensional array of log-likelihood calculations; 
        value at index `i, j` will have the value for `r_range[i]` and `q0_range[j]`

    See also
    --------
    loglik
    loglik_r_nu
    """
    if not params:
        params = dict()
    t = df.Time.unique()
    y = df.groupby('Time')['OD'].mean().to_numpy()
    y_sig = df.groupby('Time')['OD'].std().to_numpy()
    
    output = np.empty((len(r_range), len(q0_range)))
    for i, r in enumerate(r_range):
        params['r'] = r
        for j, q0 in enumerate(q0_range):
            params['q0'] = q0
            output[i,j] = loglik(t, y, y_sig, f, penalty, **params)
    
    return output


def plot_loglik(Ls, xrange, yrange, xlabel=None, ylabel=None, columns=4, fig_title=None, normalize=True,
                ax_titles=None, cmap='viridis', colorbar=True, ax_width=4, ax_height=4, ax=None):
    r"""Plots one or more log-likelihood surfaces.

    Parameters
    ----------
    Ls : sequence of numpy.ndarray
        list or tuple of log-likelihood two-dimensional arrays; if one array is given it will be converted to a size 1 list
    xrange, yrange : np.ndarray
        values on x-axis and y-axis of the plot (rows and columns of `Ls`, respectively)
    xlabel, ylabel : str, optional
        strings for x and y labels
    columns : int, optional
        number of columns in case that `Ls` has more than one matrice
    fig_title : str, optional
        a title for the whole figure
    normalize : bool, optional
        if :py:const:`True`, all matrices will be plotted using a single color scale
    ax_titles : list or tuple of str, optional
        titles corresponding to the different matrices in `Ls`
    cmap : str. optional
        name of a matplotlib colormap (to see list, call :py:func:`matplotlib.pyplot.colormaps`), defaults to `viridis`
    colorbar : bool, optional
        if :py:const:`True` a colorbar will be added to the plot
    ax_width, ax_height : int
        width and height of each panel (one for each matrice in `Ls`)
    ax : matplotlib axes or numpy.ndarray of axes
        if given, will plot into `ax`, otherwise will create a new figure

    Returns
    -------
    fig : matplotlib.figure.Figure
        figure object
    ax : numpy.ndarray
        array of axis objects

    Examples
    --------
    >>> L = loglik_r_nu(rs, nus, df, y0=y0, K=K, q0=q0, v=v)
    >>> plot_loglik(L0, rs, nus, normalize=False, fig_title=fig_title, xlabel=r'$r$', ylabel=r'$\nu$', colorbar=False)
    """
    if not isinstance(Ls, (list, tuple)):
        Ls = [Ls]
    columns = min(columns, len(Ls))
    rows = int(np.ceil(len(Ls) / columns))
    if ax is None:
        fig, ax = plt.subplots(rows, columns, sharex=True, sharey=True, figsize=(ax_width*columns, ax_height*rows))
    else:
        fig = ax.figure
    if not hasattr(ax, '__iter__'):
        ax = np.array(ax, ndmin=2)
    if ax.ndim == 1:
        ax.resize((rows, columns))
    vmin = np.nanmin(Ls)
    vmax = np.nanmax(Ls)
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
    
    for i, L in enumerate(Ls):
        row = i // columns
        col = i % columns
        _ax = ax[row, col]
        
        im = _ax.imshow(L.T, cmap=cmap, aspect=1, origin='lower')
        if normalize:
            im.set_norm(norm)
    
        if colorbar:
            plt.colorbar(im, ax=_ax, label='Log Likelihood', fraction=0.03, pad=0.25, format='%1.e')

        _ax.xaxis.grid(color='k', ls='--', alpha=0.5)
        _ax.yaxis.grid(color='k', ls='--', alpha=0.5)

        xstep = len(xrange) // 5
        xticks = list(range(0, len(xrange), xstep)) + [len(xrange)-1]
        xticklabels = xrange[xticks]
        _ax.set_xticks(xticks)
        _ax.set_xticklabels(['{:.2g}'.format(x) for x in xticklabels], rotation='vertical')

        ystep = len(xrange) // 5
        yticks = list(range(0, len(yrange), ystep)) + [len(yrange)-1]
        yticklabels = yrange[yticks]
        _ax.set_yticks(yticks)
        _ax.set_yticklabels(['{:.2g}'.format(y) for y in yticklabels])

        imax, jmax = (L == np.nanmax(L)).nonzero()
        _ax.scatter(imax, jmax, marker='o', s=50, color='r')
        
        if row == rows-1 and xlabel:
            _ax.set_xlabel(xlabel)
        if col == 0 and ylabel:
            _ax.set_ylabel(ylabel)
        if ax_titles:
            _ax.set_title(ax_titles[i])

    if fig_title:
        fig.text(0.5, 1, fig_title, fontsize=24, horizontalalignment='center')     
    fig.tight_layout()
    sns.despine()
    return fig, ax


def plot_model_loglik(m, df, fig_title=None):
    r"""Plot the log-ikelihood surfaces for :math:`\nu` over :math:`r` and :math:`q_0` over :math:`r` for given data and model fitting result.

    Parameters
    ----------
    m : lmfit.model.ModelResult
        model for which to plot the log-likelihood surface
    df : pandas.DataFrame
        data frame with `Time` and `OD` columns used to fit the model
    fig_title : str
        title for the plot

    Returns
    -------
    fig : matplotlib.figure.Figure
        figure object
    ax : numpy.ndarray
        array of axis objects

    Examples
    --------
    >>> m = curveball.models.fit_model(df)
    >>> curveball.likelihood.plot_model_loglik(m, df)

    """
    K = m.best_values['K']
    y0 = m.best_values['y0']
    r = m.best_values['r']
    nu = m.best_values.get('nu', 1)
    q0 = m.best_values.get('q0', np.inf)
    v = m.best_values.get('v', np.inf)

    rs = np.logspace(-2, 4, 100)
    nus = np.logspace(-3, 2.5, 100)
    q0s = np.logspace(-3, 2.5, 100)

    ir = ( abs(rs - r).argmin() )
    inu = ( abs(nus - nu).argmin() )
    iq0 = ( abs(q0s - m.best_values.get('q0', np.inf)).argmin() )
    ir_in = ( abs(rs - m.init_values['r']).argmin() )
    inu_in = ( abs(nus - m.init_values.get('nu', 1)).argmin() )
    iq0_in = ( abs(q0s - m.init_values.get('q0', np.inf)).argmin() )

    L0 = loglik_r_nu(rs, nus, df, y0=y0, K=K, q0=q0, v=v)
    L1 = loglik_r_q0(rs, q0s, df, y0=y0, K=K, nu=nu, v=v)
        
    w, h= plt.rcParams['figure.figsize']
    fig, ax = plt.subplots(1, 2, figsize=(w * 2, h))

    plot_loglik(L0, rs, nus, normalize=False, fig_title=fig_title,
                xlabel=r'$r$', ylabel=r'$\nu$', colorbar=False, ax=ax[0])
    ax[0].scatter(ir, inu, marker='s', s=30, color='w', edgecolors='k')
    ax[0].scatter(ir_in, inu_in, marker='^', s=30, color='w', edgecolors='k')

    plot_loglik(L1, rs, q0s, normalize=False,
                xlabel=r'$r$', ylabel=r'$q_0$', colorbar=True, ax=ax[1])
    ax[1].scatter(ir, iq0, marker='s', s=30, color='w', edgecolors='k')
    ax[1].scatter(ir_in, iq0_in, marker='^', s=30, color='w', edgecolors='k')

    fig.tight_layout()
    sns.despine()
    return fig, ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>

from setuptools import setup, find_packages
import os
import versioneer

with open('README.md') as f:
    README = f.read()

setup(
    name='curveball',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Predicting competition results from growth curves',
    long_description=README,
    long_description_content_type='text/markdown',
    keywords='microbiology biomath evolution',
    author='Yoav Ram',
    author_email='yoav@yoavram.com',
    url='https://github.com/yoavram/curveball',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',                
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: OS Independent',
    ],
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'data': ['*'],
        'plate_templates': ['*'],
    },
    data_files=[('data', ['data/Tecan_280715.xlsx', 'data/Tecan_210115.xlsx', 'data/Tecan_210115.csv', 'data/20130211_dh.zip', 'data/plate_9_OD.mat', 'data/Sunrise_180515_0916.xlsx', 'data/BioTekSynergy.xlsx']),
                  ('plate_templates', ['plate_templates/checkerboard.csv', 'plate_templates/G-RG-R.csv']),
    ],
    install_requires=[
        # remember to use 'package-name>=x.y.z,<x.y+1.0' notation (this way you get bugfixes)
        'future>=0.18.0',
        'click>=7.0',
        'lxml>=4.7.0',
        'xlrd>=1.2.0,<2.0',
        'numpy>=1.20.0',
        'scipy>=1.6.0',
        'matplotlib>=3.2.0',
        'pandas>=1.3.0',
        'seaborn>=0.11.0',
        'scikit-learn>=0.24.0',
        'sympy>=1.9.0',
        'lmfit>=1.0.0',
        'webcolors>=1.7',
        'python-dateutil>=2.8.0'
    ],
    extras_require={
        'tests': [
            'pytest',
            'pytest-cov',
            'pillow'
        ],
        'docs': [
            'sphinx>=1.3.0',
            'numpydoc',
            'sphinx_rtd_theme'
        ]
    },
    entry_points={
        'console_scripts': [
            # add cli scripts here in this form:
            'curveball=curveball.scripts.cli:cli',
        ],
    },
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>

from __future__ import absolute_import
import warnings
# catch warning caused by seaborn v0.6.0
warnings.filterwarnings(action="ignore", message="axes.color_cycle is deprecated and replaced with axes.prop_cycle; please use the latter.")
# catch warning caused by seaborn v0.7
warnings.filterwarnings(action="ignore", message="The `IPython.html` package has been deprecated. You should import from `notebook` instead. `IPython.html.widgets` has moved to `ipywidgets`.")

__license__ = u'MIT'
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from . import ioutils
from . import plots
from . import models
from . import competitions
from . import likelihood

__citation_bibtex__ = u"""@article{
Ram2015,
author = {Ram, Yoav, Dellus-Gur, Eynat, Obolski, Uri, Bibi, Maayan, Berman, Judith, and Hadany, Lilach},
doi = {10.1101/022640},						
journal = {bioRxiv},
keywords = {experimental,fitness,mathematical model,selection coefficient},
month = jul,
title = {{Predicting competition results from growth curves}},
url = {http://biorxiv.org/lookup/doi/10.1101/022640},
year = {2015}
}"""
__citation__ = u'Ram, Yoav, Eynat Dellus-Gur, Uri Obolski, Maayan Bibi, Judith Berman, and Lilach Hadany. 2015. "Predicting Competition Results from Growth Curves." bioRxiv (July 23). doi:10.1101/022640.'




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
from __future__ import division
from builtins import str
from past.utils import old_div
from unittest import TestCase, main

import sys
import os
import types
import shutil
from scipy.integrate import odeint
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from lmfit.model import Model, ModelResult
import matplotlib
from PIL import Image
import tempfile
import curveball


CI = os.environ.get('CI', 'false').lower() == 'true'
RANDOM_SEED = int(os.environ.get('RANDOM_SEED', 0))
REPS = 30
NOISE_STD = 0.04


def check_image(filename):      
	im = Image.open(filename)
	bands = im.split()
	return not all(band.getextrema() == (255, 255) for band in bands)


def logistic_ode(y, t, K, r, nu, q0, v):
	return r * y * (1 - y/K)


def richards_ode(y, t, K, r, nu, q0, v):
	return r * y * (1 - (y/K)**nu)


def baranyi_roberts_ode(y, t, K, r, nu, q0, v):
	alfa = q0 / (q0 + np.exp(-v * t))
	return alfa * r * y * (1 - (y/K)**nu)


def compare_curves(y1, y2):
	return (abs(y1 - y2) / y1).mean()


def relative_error(exp, obs):
	return abs(exp - obs) / exp


def mean_residual(model_result):
	return (np.abs(model_result.data - model_result.best_fit)).mean()


class FunctionsTestCase(TestCase):
	_multiprocess_can_split_ = True


	def test_logistic(self):
		y0=0.1; r=0.75; K=1.0
		t = np.linspace(0,12)
		y_curve = curveball.baranyi_roberts_model.baranyi_roberts_function(t, y0, r, K, 1, np.inf, np.inf)
		y_ode = odeint(logistic_ode, y0, t, args=(r, K, 1, np.inf, np.inf))
		y_ode.resize((len(t),))
		err = compare_curves(y_ode, y_curve)
		self.assertTrue(err < 1e-6)
	
	
	def test_richards(self):
		y0=0.1; r=0.75; K=1.0; nu=0.5
		t = np.linspace(0,12)
		y_curve = curveball.baranyi_roberts_model.baranyi_roberts_function(t, y0, r, K, nu, np.inf, np.inf)
		y_ode = odeint(richards_ode, y0, t, args=(r, K, nu, np.inf, np.inf))
		y_ode.resize((len(t),))
		err = compare_curves(y_ode, y_curve)
		self.assertTrue(err < 1e-6)
	
	
	def test_baranyi_roberts(self):
		y0=0.1; r=0.75; K=1.0; nu=0.5; q0=0.1; v=0.1
		t = np.linspace(0,12)
		y_curve = curveball.baranyi_roberts_model.baranyi_roberts_function(t, y0, r, K, nu, q0, v)
		y_ode = odeint(baranyi_roberts_ode, y0, t, args=(r, K, nu, q0, v))
		y_ode.resize((len(t),))
		err = compare_curves(y_ode, y_curve)
		self.assertTrue(err < 1e-6)


class ModelSelectionTestCase(TestCase):
	_multiprocess_can_split_ = True


	def setUp(self):
		if CI:
			self.folder = tempfile.mkdtemp()
		else:
			self.folder = '.'


	def tearDown(self):
		plt.close("all")
		if CI:
			shutil.rmtree(self.folder)


	def test_fit_model_logistic(self):
		df = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)
		models,fig,ax = curveball.models.fit_model(df, PLOT=True, PRINT=True)
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))
		self.assertIsNotNone(models)
		for mod in models:
			self.assertIsInstance(mod, ModelResult)
		self.assertTrue(mean_residual(models[0]) < NOISE_STD)
		

	def test_fit_model_logistic_with_param_min(self):
		df = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)       
		models,fig,ax = curveball.models.fit_model(df, PLOT=True, PRINT=True, param_min={'y0': 0.2})
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))
		self.assertIsNotNone(models)
		for mod in models:
			self.assertIsInstance(mod, ModelResult)
		self.assertTrue(models[0].best_values['y0'] >= 0.2)


	def test_fit_model_logistic_with_param_max(self):
		df = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)
		models,fig,ax = curveball.models.fit_model(df, PLOT=True, PRINT=True, param_max={'K': 0.5})
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))
		self.assertIsNotNone(models)
		for mod in models:
			self.assertIsInstance(mod, ModelResult)
		self.assertTrue(models[0].best_values['K'] <= 0.5)


	def test_fit_model_logistic_with_param_fix(self):
		df = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)
		models,fig,ax = curveball.models.fit_model(df, PLOT=True, PRINT=True, param_guess={'K': 1}, param_fix={'K'})
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))
		self.assertIsNotNone(models)
		for mod in models:
			self.assertIsInstance(mod, ModelResult)
		self.assertEqual(models[0].nvarys, 2) # one less param as we fixed K
		self.assertEqual(models[0].best_values['K'], 1)
		self.assertEqual(models[0].best_values.get('nu', 1), 1)
		self.assertEqual(models[0].best_values.get('v', np.inf), np.inf)
		self.assertEqual(models[0].best_values.get('q0', np.inf), np.inf)
		self.assertTrue(mean_residual(models[0]) < NOISE_STD)


	def test_fit_model_logistic_single_rep(self):
		df = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1, reps=1, noise_std=NOISE_STD, random_seed=RANDOM_SEED)
		models,fig,ax = curveball.models.fit_model(df, PLOT=True, PRINT=True)
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))
		self.assertIsNotNone(models)
		for mod in models:
			self.assertIsInstance(mod, ModelResult)
		self.assertTrue(mean_residual(models[0]) < NOISE_STD)



	def test_fit_model_richards_nu_05(self):
		df = curveball.models.randomize(t=24, y0=0.1, K=1, r=0.75, nu=0.5, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)
		models,fig,ax = curveball.models.fit_model(df, PLOT=True, PRINT=True)
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))
		self.assertIsNotNone(models)
		for mod in models:
			self.assertIsInstance(mod, ModelResult)
		self.assertTrue(mean_residual(models[0]) < NOISE_STD)


	def test_fit_model_logistic_lag(self):        
		df = curveball.models.randomize(t=48, y0=0.1, K=1, r=0.75, nu=1, q0=0.1, v=0.1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)
		models,fig,ax = curveball.models.fit_model(df, PLOT=True, PRINT=True)
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))    
		self.assertIsNotNone(models)
		for mod in models:
			self.assertIsInstance(mod, ModelResult)
		self.assertTrue(mean_residual(models[0]) < NOISE_STD)


	def test_fit_model_baranyi_roberts_nu_5(self):        
		df = curveball.models.randomize(t=24, y0=0.1, K=1, r=0.75, nu=5, q0=0.1, v=0.1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)
		models,fig,ax = curveball.models.fit_model(df, PLOT=True, PRINT=True)
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))    
		self.assertIsNotNone(models)
		for mod in models:
			self.assertIsInstance(mod, ModelResult)
		self.assertTrue(mean_residual(models[0]) < NOISE_STD)


class FindKTestCase(TestCase):
	_multiprocess_can_split_ = True

	def setUp(self):
		if CI:
			self.folder = tempfile.mkdtemp()
		else:
			self.folder = '.'


	def tearDown(self):
		plt.close("all")
		if CI:
			shutil.rmtree(self.folder)

	
	def test_find_K_ci_logistic(self):
		df = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED, as_df=True)
		model = curveball.baranyi_roberts_model.Logistic()
		params = model.guess(data=df.OD, t=df.Time)
		model_fit = model.fit(data=df.OD, t=df.Time, params=params)
		param_samples = curveball.models.bootstrap_params(df, model_fit, 100)
		K_est = model_fit.params['K'].value
		K_low, K_high = curveball.models.find_K_ci(param_samples)
		self.assertTrue(K_low < K_est < K_high, "K is {0}, K CI is ({1},{2})".format(K_est, K_low, K_high))


class FindLagTestCase(TestCase):
	_multiprocess_can_split_ = True

	def setUp(self):
		if CI:
			self.folder = tempfile.mkdtemp()
		else:
			self.folder = '.'


	def tearDown(self):
		plt.close("all")
		if CI:
			shutil.rmtree(self.folder)


	def test_find_lag_logistic(self):
		t, y = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED, as_df=False)
		model = curveball.baranyi_roberts_model.Logistic()
		params = model.guess(data=y, t=t)
		result = model.fit(data=y, t=t, params=params)
		lam = curveball.models.find_lag(result)
		self.assertTrue(lam < 1, "Lambda is " + str(lam))


	def test_find_lag_ci_logistic(self):
		df = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED, as_df=True)
		model = curveball.baranyi_roberts_model.Logistic()
		params = model.guess(data=df.OD, t=df.Time)
		model_fit = model.fit(data=df.OD, t=df.Time, params=params)
		param_samples = curveball.models.bootstrap_params(df, model_fit, 100)
		lam_est = curveball.models.find_lag(model_fit)
		lam_low, lam_high = curveball.models.find_lag_ci(model_fit, param_samples)
		self.assertTrue(lam_low < lam_est < lam_high, "Lambda is {0}, Lambda CI is ({1},{2})".format(lam_est, lam_low, lam_high))


	def test_find_lag_richards(self):
		for nu in [1.0, 2.0, 5.0]:
			t, y = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=nu, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED, as_df=False)
			model = curveball.baranyi_roberts_model.Richards()
			params = model.guess(data=y, t=t)
			result = model.fit(data=y, t=t, params=params)
			lam = curveball.models.find_lag(result)
			self.assertTrue(lam < 1, "Lambda is " + str(lam))
	

	def test_find_lag_baranyi_roberts(self):
		r = 0.75
		v = r
		for nu in [1.0, 2.0, 5.0]:
			for _lam in [2., 3., 4.]:
				q0 = 1.0 / (np.exp(_lam * v) - 1.0)
				t, y = curveball.models.randomize(t=16, y0=0.1, K=1, r=r, nu=nu, q0=q0, v=r, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED, as_df=False)
				model = curveball.baranyi_roberts_model.BaranyiRoberts()
				params = model.guess(data=y, t=t)
				result = model.fit(data=y, t=t, params=params)
				lam = curveball.models.find_lag(result)			      
				self.assertTrue((_lam + 1) > lam > (_lam - 1), "Lambda is " + str(lam) + " but should be " + str(_lam))

	### fails because find_lag_ci depends on sample_params which fails when covar is weird?
	# def test_find_lag_ci_baranyi_roberts(self):        
	# 	t = np.linspace(0, 24)
	# 	y0=0.1; r=0.75; K=1.0
	# 	v = r
	# 	for nu in [1.0, 2.0, 5.0]:
	# 		for _lam in [2., 3., 4.]:
	# 			q0 = 1.0 / (np.exp(_lam * v) - 1.0)
	# 			df = curveball.models.randomize(baranyi_roberts_ode, t=t, y0=y0, r=r, K=K, nu=nu, q0=q0, v=v, reps=1)
	# 			t = df.Time.as_matrix()
	# 			y = df.OD.as_matrix()
	# 			model = curveball.baranyi_roberts_model.BaranyiRobertsModel()
	# 			params = model.guess(data=y, t=t)
	# 			result = model.fit(data=y, t=t, params=params)
	# 			print('nu={}, lam={}'.format(nu, _lam))
	# 			print(result.fit_report())				
	# 			lam = curveball.models.find_lag(result)
	# 			print('lam:', lam)
	# 			lam_low,lam_high = curveball.models.find_lag_ci(result, nsamples=1000)  
	# 			self.assertTrue(lam_low < lam < lam_high, "Lambda is {2}, Lambda CI is ({0},{1})".format(lam, lam_low, lam_high))


class FindDoublingTimeTestCase(TestCase):
	_multiprocess_can_split_ = True

	def setUp(self):
		if CI:
			self.folder = tempfile.mkdtemp()
		else:
			self.folder = '.'


	def tearDown(self):
		plt.close("all")
		if CI:
			shutil.rmtree(self.folder)


	def test_find_doubling_time_logistic(self):
		t, y = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED, as_df=False)
		model = curveball.baranyi_roberts_model.Logistic()
		params = model.guess(data=y, t=t)
		result = model.fit(data=y, t=t, params=params)
		dbl = curveball.models.find_min_doubling_time(result)
		self.assertTrue(dbl > 1, "Doubling time is " + str(dbl))


	def test_find_doubling_time_ci_logistic(self):
		df = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED, as_df=True)
		model = curveball.baranyi_roberts_model.Logistic()
		params = model.guess(data=df.OD, t=df.Time)
		model_fit = model.fit(data=df.OD, t=df.Time, params=params)
		est = curveball.models.find_min_doubling_time(model_fit)
		param_samples = curveball.models.bootstrap_params(df, model_fit, 100)
		low, high = curveball.models.find_min_doubling_time_ci(model_fit, param_samples)
		self.assertTrue(low < est < high, "Doubling time is {2}, CI is ({0},{1})".format(est, low, high))


class FindMaxGrowthTestCase(TestCase):
	_multiprocess_can_split_ = True

	def setUp(self):
		if CI:
			self.folder = tempfile.mkdtemp()
		else:
			self.folder = '.'


	def tearDown(self):
		plt.close("all")
		if CI:
			shutil.rmtree(self.folder)


	def test_find_max_growth_logistic(self):
		y0 = 0.1
		K = 1.0
		r = 0.75
		df = curveball.models.randomize(t=12, y0=y0, K=K, r=r, nu=1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)
		model = curveball.baranyi_roberts_model.Logistic()
		model_fit = model.fit(df.OD, t=df.Time, y0=y0, K=K, r=r)
		  
		t1,y1,a,t2,y2,mu = curveball.models.find_max_growth(model_fit)

		self.assertTrue(relative_error(K / 2.0, y1) < 1, "y1=%.4g, K/2=%.4g" % (y1, K / 2.0))
		self.assertTrue(relative_error(K * r / 4.0, a) < 1, "a=%.4g, Kr/4=%.4g" % (a, K * r / 4.0))
		self.assertTrue(relative_error(y0, y2) < 1, "y2=%.4g, y0=%.4g" % (y2, y0))
		self.assertTrue(relative_error(r * (1 - y0 / K), mu) < 1, "mu=%.4g, r(1-y0/K)=%.4g" % (mu, r * (1 - y0 / K)))


	def test_find_max_growth_ci_logistic(self):
		y0 = 0.1
		K = 1.0
		r = 0.75
		df = curveball.models.randomize(t=12, y0=y0, K=K, r=r, nu=1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)
		model = curveball.baranyi_roberts_model.Logistic()
		model_fit = model.fit(df.OD, t=df.Time, y0=y0, K=K, r=r)
		  
		param_samples = curveball.models.bootstrap_params(df, model_fit, 100)
		_, _, a_est, _, _, mu_est = curveball.models.find_max_growth(model_fit)
		a_low, a_high, mu_low, mu_high = curveball.models.find_max_growth_ci(model_fit, param_samples)
		self.assertTrue(a_low < a_est < a_high, "a is {2}, a CI is ({0},{1})".format(a_low, a_high, a_est))
		self.assertTrue(mu_low < mu_est < mu_high, "mu is {2}, mu CI is ({0},{1})".format(mu_low, mu_high, mu_est))


	def test_find_max_growth_logistic_lag(self):
		y0 = 0.1
		K = 1.0
		r = 0.75
		lam = 3.0
		v = r
		q0 = 1.0 /(np.exp(lam * v) - 1)		
		df = curveball.models.randomize(t=12, y0=y0, K=K, r=r, nu=1, q0=q0, v=v, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)
		model = curveball.baranyi_roberts_model.LogisticLag2()		
		model_fit = model.fit(df.OD, t=df.Time, y0=y0, K=K, r=r, q0=q0, v=v)        
			 
		t1,y1,a,t2,y2,mu = curveball.models.find_max_growth(model_fit)
		 
		self.assertTrue(K > y1 > old_div(K, 2), "y1=%.4g, K/2=%.4g" % (y1, old_div(K, 2)))
		self.assertTrue(K * r / 8 < a < K * r / 4, "a=%.4g, Kr/4=%.4g" % (a, K * r / 4))
		self.assertTrue(y0 < y2 < y1, "y0=%.4g, y1=%.4g, y2=%.4g," % (y0, y1, y2))
		self.assertTrue(0 < t2 < t1, "t1=%.4g, t2=%.4g," % (t1, t2))
		self.assertTrue(a < mu < r * (1 - old_div(y0,K)), "a = %.4g, mu=%.4g, r(1-y0/K)=%.4g" % (a, mu, r * (1-old_div(y0,K))))


	def test_find_max_growth_richards(self):
		y0 = 0.1
		K = 1
		r = 0.75
		nu = 0.5
		t, y = curveball.models.randomize(t=12, y0=y0, K=K, r=r, nu=nu, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED, as_df=False)
		model = curveball.baranyi_roberts_model.Richards()
		model_fit = model.fit(data=y, t=t, y0=y0, K=K, r=r, nu=nu)
			 
		t1,y1,a,t2,y2,mu = curveball.models.find_max_growth(model_fit)
		    
		exp_y1 = K * (nu + 1)**(-1.0 / nu)
		self.assertTrue(relative_error(exp_y1, y1) < 1, "y1=%.4g, K/(nu+1)**(1/nu)=%.4g" % (y1, exp_y1))
		exp_a = r * K * nu * (nu + 1)**(- 1 - 1.0 / nu)
		self.assertTrue(relative_error(exp_a, a) < 1, "a=%.4g, rKnu/(nu+1)**(1+1/nu)=%.4g" % (a, exp_a))
		self.assertTrue(relative_error(y0, y2) < 1, "y2=%.4g, y0=%.4g" % (y2, y0))
		self.assertTrue(relative_error(r * (1 - (y0 / K)**nu), mu) < 1, "mu=%.4g, r(1-(y0/K)**nu)=%.4g" % (mu, r * (1 - (y0 / K)**nu)))


class LRTestTestCase(TestCase):
	_multiprocess_can_split_ = True

	def setUp(self):
		if CI:
			self.folder = tempfile.mkdtemp()
		else:
			self.folder = '.'


	def tearDown(self):
		plt.close("all")
		if CI:
			shutil.rmtree(self.folder)


	def test_lrtest(self):
		rng = np.random.RandomState(RANDOM_SEED)
		a,b = 1,1
		a_init,b_init = 2,1
	
		alfa = 0.05
		noise = 0.03
		t = np.linspace(0,12)
		def f(t, a, b): 
			return b + np.exp(-a * t)
		y = f(t,a,b) + rng.normal(0, noise, len(t))
		model = Model(f)
		params = model.make_params(a=a_init, b=b_init)
	
		two_var_fit = model.fit(y, t=t, params=params)
	
		params['a'].set(vary=False)
		params['b'].set(vary=True)
		one_var_fit = model.fit(y, t=t, params=params)
	
		prefer_m1,pval,D,ddf = curveball.models.lrtest(one_var_fit, two_var_fit, alfa)
		self.assertTrue(prefer_m1)


	def test_has_lag_logistic(self):
		df = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)

		models = curveball.models.fit_model(df, PLOT=False, PRINT=True)
		lag = curveball.models.has_lag(models)
		self.assertFalse(lag)


	def test_has_lag_richards(self):
		df = curveball.models.randomize(t=6, y0=0.1, K=1, r=0.75, nu=5.0, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)

		models,fig,ax = curveball.models.fit_model(df, PLOT=True, PRINT=True)
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))   
		lag = curveball.models.has_lag(models)
		self.assertFalse(lag)


	def test_has_lag_baranyi_roberts(self):
		df = curveball.models.randomize(t=24, y0=0.1, K=1, r=0.75, nu=5.0, q0=0.1, v=0.1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)
		
		models,fig,ax = curveball.models.fit_model(df, PLOT=True, PRINT=True)
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))
		lag = curveball.models.has_lag(models)
		self.assertTrue(lag)


	def test_has_nu_logistic(self):
		df = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)

		models = curveball.models.fit_model(df, PLOT=False, PRINT=True)
		result = curveball.models.has_nu(models)
		self.assertFalse(result)

	
	def test_has_nu_richards(self):
		df = curveball.models.randomize(t=6, y0=0.1, K=1, r=0.75, nu=5.0, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)

		models, fig, ax = curveball.models.fit_model(df, PLOT=True, PRINT=True)
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))
		result = curveball.models.has_nu(models)
		self.assertTrue(result)

	## FIXME this is too hard, nu easiliy compensates for lag params

	# def test_has_nu_logistic_lag(self):
	# 	df = curveball.models.randomize(t=36, y0=0.1, K=1, r=0.75, nu=1, q0=0.1, v=0.1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)

	# 	models, fig, ax = curveball.models.fit_model(df, PLOT=True, PRINT=True)
	# 	self.assertIsInstance(fig, matplotlib.figure.Figure)
	# 	filename = sys._getframe().f_code.co_name + ".png"
	# 	fig.savefig(filename)
	# 	self.assertTrue(check_image(filename))
	# 	result = curveball.models.has_nu(models)
	# 	self.assertFalse(result)

	# def test_has_nu_baranyi_roberts_nu_1(self):
	# 	df = curveball.models.randomize(t=32, y0=0.1, K=1, r=0.75, nu=1.0, q0=0.1, v=0.1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)
		
	# 	models,fig,ax = curveball.models.fit_model(df, PLOT=True, PRINT=True)
	# 	self.assertIsInstance(fig, matplotlib.figure.Figure)
	# 	filename = sys._getframe().f_code.co_name + ".png"
	# 	fig.savefig(filename)
	# 	self.assertTrue(check_image(filename))
	# 	result = curveball.models.has_nu(models, PRINT=True)
	# 	self.assertFalse(result)


	# def test_has_nu_baranyi_roberts_nu_01(self):
	# 	df = curveball.models.randomize(t=120, y0=0.1, K=1, r=0.75, nu=0.1, q0=0.1, v=0.1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)		
		
	# 	models,fig,ax = curveball.models.fit_model(df, PLOT=True, PRINT=True)
	# 	self.assertIsInstance(fig, matplotlib.figure.Figure)
	# 	filename = sys._getframe().f_code.co_name + ".png"
	# 	fig.savefig(filename)
	# 	self.assertTrue(check_image(filename))       
	# 	result = curveball.models.has_nu(models, PRINT=True)
	# 	self.assertTrue(result)


	def test_has_nu_baranyi_roberts_nu_5(self):
		df = curveball.models.randomize(t=32, y0=0.1, K=1, r=0.75, nu=5.0, q0=0.1, v=0.1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)

		models,fig,ax = curveball.models.fit_model(df, PLOT=True, PRINT=True)
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))
		result = curveball.models.has_nu(models, PRINT=True)
		self.assertTrue(result)


class BenchmarkTestCase(TestCase):
	_multiprocess_can_split_ = True

	def setUp(self):
		if CI:
			self.folder = tempfile.mkdtemp()
		else:
			self.folder = '.'


	def tearDown(self):
		plt.close("all")
		if CI:
			shutil.rmtree(self.folder)


class OutliersTestCase(TestCase):
	_multiprocess_can_split_ = True


	def setUp(self):
		self.filename = os.path.join("data", "Tecan_210115.csv")
		self.df = pd.read_csv(self.filename)
		self.df = self.df[self.df.Strain == 'R']
		self.model_fit = curveball.models.fit_model(self.df, PLOT=False, PRINT=False)[0]
		if CI:
			self.folder = tempfile.mkdtemp()
		else:
			self.folder = '.'


	def tearDown(self):
		plt.close("all")
		if CI:
			shutil.rmtree(self.folder)


	def test_cooks_distance(self):
		D = curveball.models.cooks_distance(self.df, self.model_fit)
		self.assertEqual(set(D.keys()), set(self.df.Well), msg=D.keys())
		distances = np.array(list(D.values()))
		for d in distances:
			self.assertIsInstance(d, float)
		self.assertTrue( (distances > 0).all(), msg=D.values() )


	def test_find_outliers(self):		
		outliers,fig,ax = curveball.models.find_outliers(self.df, self.model_fit, PLOT=True)
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))       
		self.assertTrue(pd.Series(outliers).isin(self.df.Well).all())
		self.assertTrue(len(outliers) < len(self.df.Well.unique()))


	def test_find_all_outliers(self):        
		outliers,fig,ax = curveball.models.find_all_outliers(self.df, self.model_fit, PLOT=True)
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))       
		self.assertIsNotNone(outliers)
		self.assertTrue(len(outliers) > 0)
		for v in outliers: self.assertTrue(len(v) > 0)
		self.assertTrue(pd.Series(sum(outliers, [])).isin(self.df.Well).all())
		self.assertTrue(len(sum(outliers, [])) < len(self.df.Well.unique()))


class SamplingTestCase(TestCase):
	_multiprocess_can_split_ = True


	def test_sample_params(self):
		t, y = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1.0, reps=1, noise_std=NOISE_STD, random_seed=RANDOM_SEED, as_df=False)
		model = curveball.baranyi_roberts_model.Logistic()
		params = model.guess(data=y, t=t)
		model_fit = model.fit(data=y, t=t, params=params)
		self.assertIsNotNone(model_fit.covar)
		sample_params = curveball.models.sample_params(model_fit, 100)
		self.assertIsNotNone(sample_params)
		self.assertEqual(sample_params.shape, (100, 3))


	def test_sample_params_with_params(self):
		t, y = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1.0, reps=1, noise_std=NOISE_STD, random_seed=RANDOM_SEED, as_df=False)
		model = curveball.baranyi_roberts_model.Logistic()
		params = model.guess(data=y, t=t)
		model_fit = model.fit(data=y, t=t, params=params)
		self.assertIsNotNone(model_fit.covar)
		sample_params = curveball.models.sample_params(model_fit, 100, params={'K': 1.0})
		self.assertIsNotNone(sample_params)
		self.assertEqual(sample_params.shape, (100, 3))


	def test_sample_params_with_covar(self):
		t, y = curveball.models.randomize(t=12, y0=0.1, K=1, r=0.75, nu=1.0, reps=1, noise_std=NOISE_STD, random_seed=RANDOM_SEED, as_df=False)
		model = curveball.baranyi_roberts_model.Logistic()
		params = model.guess(data=y, t=t)
		model_fit = model.fit(data=y, t=t, params=params)
		self.assertIsNotNone(model_fit.covar)
		covar = model_fit.covar
		rng = np.random.RandomState(RANDOM_SEED)
		covar = rng.exponential(0.001, (3, 3))
		sample_params = curveball.models.sample_params(model_fit, 100, covar=covar)
		self.assertIsNotNone(sample_params)
		self.assertEqual(sample_params.shape[1], 3)
		self.assertTrue(95 <= sample_params.shape[0] <= 100)


	def test_bootstrap_params(self):
		plate = pd.read_csv('plate_templates/G-RG-R.csv')
		df = curveball.ioutils.read_tecan_xlsx('data/Tecan_280715.xlsx', plate=plate)
		model_result = curveball.models.fit_model(
			df, models=curveball.baranyi_roberts_model.BaranyiRoberts, 
			PLOT=False, PRINT=False)[0]
		sample_params = curveball.models.bootstrap_params(df, model_result, 10)
		self.assertIsInstance(sample_params, pd.DataFrame)
		self.assertEqual(sample_params.shape, (10, 6))


class GuessTestCase(TestCase):
	_multiprocess_can_split_ = True


	def test_guess_nu(self):
		df = curveball.models.randomize(t=32, y0=0.1, K=1, r=0.75, nu=1.0, reps=30, noise_std=NOISE_STD, random_seed=RANDOM_SEED)
		nu, fig, ax = curveball.baranyi_roberts_model.guess_nu(t=df.Time, N=df.OD, PLOT=True, PRINT=True)
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))               
		self.assertIsInstance(nu, float)
		self.assertTrue(nu > 0)


class WeightsTestCase(TestCase):
	_multiprocess_can_split_ = True


	def test_calc_weights(self):
		df = curveball.models.randomize(t=32, y0=0.1, K=1, r=0.75, nu=1.0, reps=30, noise_std=NOISE_STD, random_seed=RANDOM_SEED)
		weights, fig, ax = curveball.models.calc_weights(df, PLOT=True)
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		filename = sys._getframe().f_code.co_name + ".png"
		fig.savefig(filename)
		self.assertTrue(check_image(filename))
		self.assertIsNotNone(weights)		
		self.assertIsInstance(weights, np.ndarray)
		self.assertTrue(len(weights) == df.shape[0])


class IssuesTestCase(TestCase):
	'''Tests that came up from bugs and other issues.

	See `Curveball issues <https://github.com/yoavram/curveball>`_.
	'''
	_multiprocess_can_split_ = True

	## I no longer care if covar is None; I can and should use bootstrap_params instead of sample_params
	# def test_covar_exists_issue27(self):
	# 	'''`Issue 27 <https://github.com/yoavram/curveball/issues/27>`_.
	# 	'''
	# 	plate = pd.read_csv('plate_templates/G-RG-R.csv')
	# 	df = curveball.ioutils.read_tecan_xlsx('data/Tecan_280715.xlsx', plate=plate)
	# 	self.assertTrue('R' in df.Strain.unique())
	# 	df = df[df.Strain == 'R']
	# 	models_R,fig,ax = curveball.models.fit_model(df[df.Time<=16], PLOT=True, PRINT=True)
	# 	filename = sys._getframe().f_code.co_name + ".png"
	# 	fig.savefig(filename)
	# 	self.assertIsNotNone(models_R[0].covar)

	## FIXME this is too hard, nu easiliy compensates for lag params

	# def test_has_nu_issue22(self):
	# 	'''`Issue 22 <https://github.com/yoavram/curveball/issues/22>`_.
	# 	'''
	# 	df = curveball.models.randomize(t=48, y0=0.1, K=1, r=0.75, nu=5.0, q0=0.1, v=0.1, reps=REPS, noise_std=NOISE_STD, random_seed=RANDOM_SEED)

	# 	models, fig, ax = curveball.models.fit_model(df, PLOT=True, PRINT=True)
	# 	self.assertIsInstance(fig, matplotlib.figure.Figure)
	# 	filename = sys._getframe().f_code.co_name + ".png"
	# 	fig.savefig(filename)
	# 	self.assertTrue(check_image(filename))  

	# 	self.assertTrue('nu' in models[0].best_values)
	# 	nu = models[0].best_values['nu']
	# 	self.assertTrue(1.0 < nu < 10.0, nu)
	# 	has = curveball.models.has_nu(models)
	# 	self.assertTrue(has)


	def test_Dfun_works(self):
		t, y = curveball.models.randomize(as_df=False)
		for model_class in curveball.models.get_models(curveball.baranyi_roberts_model):
			model = model_class()
			params = model.guess(data=y, t=t)
			dfun = curveball.models.make_Dfun(model, params)
			self.assertTrue(hasattr(dfun, '__call__'))
			result = model.fit(data=y, t=t, params=params, fit_kws={'Dfun': dfun, 'col_deriv':True})
			self.assertIsInstance(result, ModelResult)


	def test_is_model(self):
		for model in curveball.models.get_models(curveball.baranyi_roberts_model):
			self.assertTrue(curveball.models.is_model(model))


	def test_make_Dfun(self):
		model = curveball.baranyi_roberts_model.BaranyiRoberts()
		params = model.make_params(y0=0.1, K=1, r=1, nu=1, v=1, q0=1)
		Dfun = curveball.models.make_Dfun(model, params)
		self.assertIsInstance(Dfun, types.FunctionType)

		t, y = curveball.models.randomize(as_df=False)
		res = Dfun(params, y, None, t)
		self.assertIsInstance(res, np.ndarray)
		self.assertEqual(res.shape, (len(params), len(t)))


if __name__ == '__main__':
	main()
